import { calendar_v3, google } from 'googleapis';
import { getAccessToken, getOAuthClient } from './oauthAccess';
import calendar_ids from '../database/calendar_ids';
import user_google_tokens from '../database/user_google_tokens';
import { getUser } from '../database/user';

import { cacheEventsTime, defaultUserSettings } from '../config';

import { OpenHoursItem, getClosedHours, getEndOfDay, getOpenHours, getStartOfDayOfUser } from './openHoursList';
import { MasterUser, User, UserID } from '../user/types';
import { Store, StoreID, getStore } from '../database/store';
import { getAllBusyEventsFromDatabase } from '../calendarSync/getAllEvents';
import redisClient from '../redis/init';
import MyRedisKeys from '../redis/keys';
import { storeLogger, userLogger } from '../logger/logger';

type CalendarID = string;
type timestamp = number;

type EventType = calendar_v3.Schema$Event;
type SimpleEventType = {
    start: Date;
    end: Date;
};

type SimpleUserEventType = {
    userId: string;
    events: SimpleEventType[];
};

async function getEvents(calendarID: string, earliestBookingTime: Date, endDate: Date): Promise<EventType[]> {
    try {
        const _myEvents = await getAllBusyEventsFromDatabase(calendarID, earliestBookingTime, endDate);

        // convert _myEvents to google calendar events
        let myEvents: EventType[] = [];
        for (const event of _myEvents) {
            myEvents.push({
                id: event.event_id,
                start: event.start_time ? { dateTime: event.start_time.toISOString() } : undefined,
                end: event.end_time ? { dateTime: event.end_time.toISOString() } : undefined,
                status: event.status_val,
                updated: event.updated_time ? event.updated_time.toISOString() : undefined,
                transparency: event.transparency,
            });
        }

        return myEvents;
    } catch (error) {
        console.error(error);
        throw error;
    }
}

async function getBookedTimeFromOneUser(store: Store, user: User, day: Date, time?: Date): Promise<{ start: Date; end: Date }[]> {
    return new Promise(async (resolve, reject) => {
        try {
            let openHours: OpenHoursItem[];
            let endDate: Date;
            let earliestBookingTime: Date;

            try {
                openHours = await getOpenHours(store.store_id); /* eg [
                    { start: 2024-01-10T04:30:00.000Z, end: 2024-01-10T14:00:00.000Z },
                    { start: 2024-01-11T04:30:00.000Z, end: 2024-01-11T14:00:00.000Z },
                    { start: 2024-01-12T04:30:00.000Z, end: 2024-01-12T14:00:00.000Z },
                ]*/

                // day is the day to check

                endDate = await getEndOfDay(openHours, day);
                earliestBookingTime = await getStartOfDayOfUser(user, openHours, day, time !== undefined);
            } catch (error) {
                reject(error);
                return;
            }
            // check if time is set
            if (time !== undefined) {
                if (time.getTime() < earliestBookingTime.getTime()) {
                    reject('no free time found');
                    return;
                }
            }

            if (endDate < earliestBookingTime) {
                reject('no free time found');
                return;
            }

            let masterCalendarIDs: string[] = [];
            let myCalendarIDs: string[] = [];
            if (user.settings.calendar_using_primary_calendar === true) {
                try {
                    const primaryCalendarID = (await calendar_ids.getPrimaryCalendar(user.user_id)).calendar_id;

                    myCalendarIDs.push(primaryCalendarID);
                    userLogger.debug(user.user_id, 'using primary calendar for user', user.username);
                } catch (error) {}
            }
            masterCalendarIDs.push(await calendar_ids.getEventCalendarID(user));

            async function getAllEvents(calendarIDs: string[]) {
                let events: EventType[] = [];

                for (const calendarID of calendarIDs) {
                    const calRes = await getEvents(calendarID, earliestBookingTime, endDate);
                    events.push(...calRes);
                }

                return events;
            }

            let masterEvents = await getAllEvents(masterCalendarIDs);
            let myEvents = myCalendarIDs.length >= 1 ? await getAllEvents(myCalendarIDs) : [];

            // invert openHours and only keep closed hours that are between earliestBookingTime and endTime
            const closedHours = getClosedHours(openHours, earliestBookingTime, endDate);

            // merge myEvents and masterEvents into one array and merge overlapping events into one
            let _events: EventType[] = [...masterEvents, ...myEvents];

            // change events to only contain start and end time
            let __events: (SimpleEventType | undefined)[] = _events.map((event) => {
                const start = event.start?.dateTime || event.start?.date;
                const end = event.end?.dateTime || event.end?.date;

                if (!start || !end) return;

                return { start: new Date(start), end: new Date(end) };
            });

            __events = [...__events, ...closedHours];

            // remove undefined
            let events: SimpleEventType[] = __events.filter((event) => event !== undefined) as { start: Date; end: Date }[];

            // sort events by start time
            events.sort((a, b) => a.start.getTime() - b.start.getTime());

            // merge overlapping events
            let mergedEvents: SimpleEventType[] = [];
            for (let i = 0; i < events.length; i++) {
                const event = events[i];

                let found = false;
                for (let j = 0; j < mergedEvents.length; j++) {
                    const mergedEvent = mergedEvents[j];

                    if (event.start >= mergedEvent.start && event.start <= mergedEvent.end) {
                        if (event.end > mergedEvent.end) {
                            mergedEvent.end = event.end;
                        }
                        found = true;
                        break;
                    } else if (event.end >= mergedEvent.start && event.end <= mergedEvent.end) {
                        if (event.start < mergedEvent.start) {
                            mergedEvent.start = event.start;
                        }
                        found = true;
                        break;
                    } else if (event.start <= mergedEvent.start && event.end >= mergedEvent.end) {
                        mergedEvent.start = event.start;
                        mergedEvent.end = event.end;
                        found = true;
                        break;
                    }
                }

                if (!found) {
                    mergedEvents.push(event);
                }
            }

            //console.log('mergedEvents: ', mergedEvents);
            userLogger.debug(user.user_id, 'booked time: ', JSON.stringify(mergedEvents, null, 2));

            resolve(mergedEvents); // return booked time
        } catch (error) {
            userLogger.error(user.user_id, (error as Error).toString());
            reject(error);
        }
    });
}

async function getFreeTimeFromOneUser(user: User, bookTimeLength: number, day: Date): Promise<SimpleEventType[]> {
    return new Promise(async (resolve, reject) => {
        try {
            //check if user is in redis

            let redisFreeTime = await redisClient.get(MyRedisKeys.freeTimeFromOneUserCache(user.user_id, bookTimeLength, day));
            if (redisFreeTime) {
                let freeTime: SimpleEventType[] = JSON.parse(redisFreeTime);

                // change start and end to Date
                for (const time of freeTime) {
                    time.start = new Date(time.start);
                    time.end = new Date(time.end);
                }

                resolve(freeTime);
                return;
            }
        } catch (error) {
            userLogger.error(user.user_id, (error as Error).toString());
        }

        try {
            userLogger.debug(user.user_id, 'calc freetime: ', day.toLocaleDateString());

            const store = await getStore(user.store_id);
            const openHours = await getOpenHours(store.store_id);

            // check if "day" is in user.settings.calendar_max_future_booking_days range
            const maxFutureBookingDays = user.settings.calendar_max_future_booking_days;
            const maxFutureBookingDaysDate = new Date(new Date().getTime() + maxFutureBookingDays * 24 * 60 * 60 * 1000);
            if (day.getTime() > maxFutureBookingDaysDate.getTime()) {
                reject('day is out of range');
                return;
            }

            let earliestBookingTime = await getStartOfDayOfUser(user, openHours, day);
            let bookedTime = await getBookedTimeFromOneUser(store, user, day);

            let endTime = await getEndOfDay(openHours, day);

            // create freeTime from bookedTime
            let freeTime: SimpleEventType[] = [];
            for (let i = 0; i < bookedTime.length; i++) {
                const time = bookedTime[i];

                if (i == 0) {
                    if (earliestBookingTime < time.start) {
                        freeTime.push({ start: earliestBookingTime, end: time.start });
                    }
                } else {
                    freeTime.push({ start: bookedTime[i - 1].end, end: time.start });
                }
            }

            if (bookedTime.length == 0) {
                freeTime.push({ start: earliestBookingTime, end: endTime });
            } else {
                freeTime.push({ start: bookedTime[bookedTime.length - 1].end, end: endTime });
            }

            // remove freeTime that is too short
            let _freeTime: SimpleEventType[] = [];
            for (const time of freeTime) {
                if (time.end.getTime() - time.start.getTime() + 1000 /*1000ms: just for better fitting*/ >= bookTimeLength * 60 * 1000) {
                    // set time.end to time.end - bookTimeLength
                    time.end = new Date(time.end.getTime() - bookTimeLength * 60 * 1000);

                    // check if time.end is before time.start
                    if (time.end.getTime() < time.start.getTime()) {
                        // if so, set time.end to time.start
                        time.end = new Date(time.start.getTime());
                    }

                    _freeTime.push(time);
                }
            }

            freeTime = _freeTime;

            // to redis
            try {
                await redisClient.set(MyRedisKeys.freeTimeFromOneUserCache(user.user_id, bookTimeLength, day), JSON.stringify(freeTime), { EX: 24 * 60 * 60 });
            } catch (error) {}

            resolve(freeTime);
        } catch (error) {
            try {
                await redisClient.set(MyRedisKeys.freeTimeFromOneUserCache(user.user_id, bookTimeLength, day), JSON.stringify([]), { EX: 24 * 60 * 60 });
            } catch (error) {}

            reject(error);
        }
    });
}

async function getFreeTimeOfDay(storeID: StoreID, userIDs: string[], bookTimeLength: number, day: Date): Promise<SimpleUserEventType[]> {
    return new Promise(async (resolve, reject) => {
        try {
            const redisKey = MyRedisKeys.freeTimeOfDayCache(storeID, userIDs, bookTimeLength, day);

            try {
                //check if user is in redis
                let redisFreeTime = await redisClient.get(redisKey);

                if (redisFreeTime) {
                    let freeTime: SimpleUserEventType[] = JSON.parse(redisFreeTime);

                    // change start and end to Date
                    for (const user of freeTime) {
                        for (const time of user.events) {
                            time.start = new Date(time.start);
                            time.end = new Date(time.end);
                        }
                    }

                    resolve(freeTime);
                    return;
                }
            } catch (error) {}

            try {
                const store = await getStore(storeID);
                if (store === undefined) {
                    reject('store not found');
                    return;
                }

                let users: User[] = [];
                for (const userID of userIDs) {
                    try {
                        const user = await getUser(userID);
                        if (user === undefined) {
                            reject('user not found');
                            return;
                        }
                        if (user.store_id !== storeID) {
                            //reject('user.store_id !== storeID');
                            //return;
                            continue;
                        }
                        users.push(user);
                    } catch (error) {
                        userLogger.error(userID, (error as Error).toString());
                    }
                }

                const openHours = await getOpenHours(storeID);

                // day is the day to check

                let freeTime: SimpleUserEventType[] = [];

                let endTime = await getEndOfDay(openHours, day); // this also checks if day is in openHours

                for (const user of users) {
                    try {
                        const _freeTime = await getFreeTimeFromOneUser(user, bookTimeLength, day);

                        freeTime.push({ userId: user.user_id, events: _freeTime });
                    } catch (error) {
                        userLogger.debug(user.user_id, '@reject', (error as Error).toString());
                    }
                }

                // cache to redis
                try {
                    await redisClient.set(redisKey, JSON.stringify(freeTime), { EX: 24 * 60 * 60 });
                } catch (error) {}

                resolve(freeTime);
            } catch (error) {
                reject(error);
            }
        } catch (error) {
            reject(error);
        }
    });
}

async function getFreeNextDays(store_id: StoreID, _users: UserID[], serviceDuration: number, _date: Date): Promise<Date[]> {
    return new Promise(async (resolve, reject) => {
        try {
            // get next days until end of month
            const nextDays = [];

            const date = new Date(_date); // 11.01.2024 11:17:19
            let currentDate = date.getDate(); // 1-31
            date.setHours(0, 0, 0, 0); // 11.01.2024 00:00:00
            date.setDate(1); // 01.01.2024 00:00:00
            date.setMonth(date.getMonth() + 1); // 01.02.2024 00:00:00
            date.setDate(0); // 31.01.2024 00:00:00
            const lastDayOfMonth = date.getDate();

            // set currentDate to new Date() when month of _date is the current month

            if (new Date().getMonth() === _date.getMonth() && new Date().getFullYear() === _date.getFullYear()) {
                _date = new Date();
                currentDate = _date.getDate();
            }

            for (let i = currentDate - 1; i < lastDayOfMonth; i++) {
                const day = new Date(_date);

                if (currentDate !== i + 1) {
                    day.setHours(0, 0, 0, 0);
                    day.setDate(i + 1);
                }
                nextDays.push(day);
            }

            let freeDays: Date[] = [];
            let users: UserID[] = [];

            for (const user_id of _users) {
                try {
                    const user = await getUser(user_id);

                    await calendar_ids.getEventCalendarID(user);

                    users.push(user.user_id);
                    userLogger.debug(user.user_id, 'calc getFreeNextDays:');
                } catch (error) {}
            }

            if (users.length > 0) {
                for (let i = 0; i < nextDays.length; i++) {
                    const day = nextDays[i];
                    try {
                        //const measureTime = new Date().getTime();
                        const freeTime = await getFreeTimeOfDay(store_id, users, serviceDuration, day);

                        let isDayFree = false;
                        for (const key in freeTime) {
                            const time = freeTime[key].events;
                            if (time.length > 0) {
                                isDayFree = true;
                                break;
                            }
                        }
                        if (isDayFree) {
                            freeDays.push(day);
                        }

                        //console.log('getFreeNextDays took ', new Date().getTime() - measureTime, 'ms');
                    } catch (error) {
                        //console.error(error);
                    }
                }
            }

            resolve(freeDays);
        } catch (error) {
            console.error(error);
            storeLogger.error(store_id, (error as Error).toString());
            reject(error);
        }
    });
}

function printDate(str: string, date: Date) {
    console.log(
        str,
        date.toLocaleDateString('de-DE', {
            day: '2-digit',
            month: '2-digit',
            year: 'numeric',
            hour: '2-digit',
            minute: '2-digit',
            second: '2-digit',
        })
    );
}

export { getBookedTimeFromOneUser, getFreeTimeOfDay, getFreeNextDays, EventType };
