import { OAuth2Client } from 'google-auth-library';
import { google } from 'googleapis';
import { clientId, clientSecret, defaultPrimaryCalendarId } from '../config';
import { getAccessToken, getOAuthClient } from './oauthAccess';

import lang from '../lang/default';
import calendar_ids, { CalendarType } from '../database/calendar_ids';
import { getAllUsersIDsByStoreID, getUser } from '../database/user';
import { Store, getStoreByUser } from '../database/store';
import { User } from '../user/types';
import { enableCalendarSync } from '../calendarSync/refreshSync';
import { userLogger } from '../logger/logger';

async function getPrefix(store: Store) {
    return new Promise(async (resolve, reject) => {
        try {
            resolve(store.name);
        } catch (error) {
            console.log(error);
            reject(error);
        }
    });
}

async function initCalendar(userId: string) {
    return new Promise(async (resolve, reject) => {
        function noPermission() {
            userLogger.error(userId, 'No permission to access calendar');
            reject('403');
        }

        try {
            // wait 1 second
            await new Promise((resolve) => setTimeout(resolve, 1000));
            userLogger.info(userId, 'initCalendar...');

            const user = await getUser(userId);

            const store = await getStoreByUser(user);
            const ownerUser = await getUser(store.owner_user_id);
            const accessToken = await getAccessToken(ownerUser.user_id);

            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            let calendars = await calendar_ids.getCalendars(ownerUser.user_id);

            console.log('calendars', calendars);
            userLogger.debug(userId, 'calendars', JSON.stringify(calendars, null, 2));

            // delete calendar if it not exists in google
            for (let i = 0; i < calendars.length; i++) {
                const cal = calendars[i];

                if (cal.type === 'primary') {
                    if (user.isOwner !== true) continue;

                    await enableCalendarSync(cal.calendar_id);
                    userLogger.info(userId, 'primary calendar enabled', cal.calendar_id);
                    continue;
                }
                if (user.isOwner !== true && cal.worker_user_id !== user.user_id) continue;

                try {
                    /*const googleCalendar = await calendar.calendarList.get({
                        calendarId: cal.calendar_id,
                    });*/
                    const googleCalendar = await calendar.calendars.get({
                        calendarId: cal.calendar_id,
                    });

                    await enableCalendarSync(cal.calendar_id);

                    // check if i own the googleCalendar
                    //console.log(googleCalendar);
                } catch (error) {
                    console.error('initCalendar:', (error as any).response?.status);

                    if ((error as any).response?.status == 404) {
                        delete calendars[i];
                        await calendar_ids.deleteCalendar(user, cal.calendar_id);

                        userLogger.warn(userId, 'deleted calendar ' + cal.calendar_id);
                    } else if ((error as any).response?.status == 403) {
                        noPermission();
                        return;
                    } else {
                        userLogger.error(userId, 'initCalendar:', JSON.stringify(error, null, 2));
                    }
                }
            }

            // delete all undefined values in calendars
            calendars = calendars.filter((value) => value);

            const requiredCalendars: CalendarType[] = [];

            if (user.isOwner === true) {
                requiredCalendars.push('myEvents');
                requiredCalendars.push('openHours');
            } else {
                requiredCalendars.push('workerEvents');
            }
            requiredCalendars.push('primary');

            for (let i = 0; i < requiredCalendars.length; i++) {
                const requiredCalendar = requiredCalendars[i];

                let found = false;
                let foundId = '';

                for (let j = 0; j < calendars.length; j++) {
                    const cal = calendars[j];

                    if (cal.type == requiredCalendar) {
                        if (requiredCalendar === 'workerEvents' && cal.worker_user_id !== user.user_id) continue;

                        found = true;
                        foundId = cal.calendar_id;
                        break;
                    }
                }

                async function insertCalendar() {
                    console.log('inserting calendar ' + requiredCalendar);
                    userLogger.info(userId, 'inserting calendar ' + requiredCalendar);

                    let prefix = user.username;

                    if (requiredCalendar === 'primary') {
                        await calendar_ids.addCalendar(userId, defaultPrimaryCalendarId + '#' + userId, requiredCalendar);
                        return;
                    }

                    if (requiredCalendar === 'myEvents' || requiredCalendar === 'openHours') prefix = lang.calendar[requiredCalendar];

                    const cal = await calendar.calendars.insert({
                        requestBody: {
                            summary: prefix + ' - ' + (await getPrefix(store)),
                            description: prefix,
                        },
                    });

                    if (!cal.data.id) {
                        console.error('cal.data.id is undefined');
                        userLogger.error(userId, 'cal.data.id is undefined');
                        reject('cal.data.id is undefined');
                        return;
                    }

                    await calendar_ids.addCalendar(ownerUser.user_id, cal.data.id, requiredCalendar, user.isOwner !== true ? user.user_id : undefined);
                }

                if (!found || (requiredCalendar === 'primary' && user.isOwner !== true)) {
                    await insertCalendar();
                } else {
                    try {
                        if (requiredCalendar === 'primary') continue;

                        const caltest = await calendar.calendars.get({
                            calendarId: foundId,
                        });
                        //console.log('found: ', caltest.data);
                    } catch (error) {
                        console.log(error);
                        console.log('calendar not found in init');
                        userLogger.warn(userId, 'initCalendar: calendar not found, creating...');

                        await insertCalendar();
                    }
                }
            }

            if (user.isOwner === true) {
                //also initCalendar for all workers
                try {
                    const workersIDs = await getAllUsersIDsByStoreID(store.store_id);

                    for (let i = 0; i < workersIDs.length; i++) {
                        if (workersIDs[i] === user.user_id) continue;
                        const workerID = workersIDs[i];

                        await initCalendar(workerID);
                    }
                } catch (error) {
                    console.error(error);
                }
            }

            resolve(true);
        } catch (error) {
            console.log(error);

            if ((error as any).response?.status == 403) {
                noPermission();
                return;
            }

            reject(error);
        }
    });
}

async function removeCalendar(user: User) {
    return new Promise(async (resolve, reject) => {
        try {
            const store = await getStoreByUser(user);
            const ownerUser = await getUser(store.owner_user_id);
            const accessToken = await getAccessToken(ownerUser.user_id);

            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            const calendars = await calendar_ids.getCalendars(ownerUser.user_id);

            for (let i = 0; i < calendars.length; i++) {
                const cal = calendars[i];

                if ((cal.type == 'workerEvents' && cal.worker_user_id === user.user_id) || user.isOwner === true) {
                    console.log('deleting calendar ' + cal.calendar_id);
                    console.log('From User:', user.user_id, ' (Owner:', ownerUser.user_id, ')');
                    userLogger.warn(user.user_id, 'deleting calendar ' + cal.calendar_id, 'From User:', user.user_id, ' (Owner:', ownerUser.user_id, ')');

                    /*await calendar.calendars.delete({
                        calendarId: cal.calendar_id,
                    });*/ // no permission to delete calendar

                    await calendar_ids.deleteCalendar(user, cal.calendar_id);
                }
            }

            resolve(true);
        } catch (error) {
            console.log(error);
            reject(error);
        }
    });
}

export default initCalendar;
export { removeCalendar };
