import { google } from 'googleapis';
import { CalendarEvent } from '../database/reservedCustomer';
import { Store, getStoreByUser } from '../database/store';
import { getUser } from '../database/user';
import { User } from '../user/types';
import { EventType, getBookedTimeFromOneUser } from './getFreeTime';
import { getAccessToken, getOAuthClient } from './oauthAccess';
import { get } from 'http';
import { userInfo } from 'os';
import calendar_ids from '../database/calendar_ids';
import { getActivity } from '../database/services';
import { getConnection } from '../database/initDatabase';

import { insertEvent as insertEventGoogleMariadb, deleteEvent as deleteEventGoogleMariadb } from '../calendarSync/misc';
import clearRedisCache from '../redis/clearCache';
import { userLogger } from '../logger/logger';

async function checkIfTimeIsAvailable(store: Store, user: User, start: Date, end: Date): Promise<void> {
    return new Promise(async (resolve, reject) => {
        try {
            // check if start is before end
            if (start >= end) {
                reject('Start time is after end time');
                return;
            }

            // check if start and end is in the same day, month and year
            if (start.getDate() !== end.getDate() || start.getMonth() !== end.getMonth() || start.getFullYear() !== end.getFullYear()) {
                reject('Start and end is not in the same day, month and year');
                return;
            }

            await clearRedisCache(store.store_id);

            const bookedTime = await getBookedTimeFromOneUser(store, user, start, start /* time is important */);
            console.log('bookedTime', bookedTime);

            // check if start and end is not overlapping with other events in bookedTime
            for (let i = 0; i < bookedTime.length; i++) {
                if (start >= bookedTime[i].start && start < bookedTime[i].end) {
                    reject('Start time is overlapping with another event');
                    return;
                }
                if (end > bookedTime[i].start && end <= bookedTime[i].end) {
                    reject('End time is overlapping with another event');
                    return;
                }

                if (start <= bookedTime[i].start && end >= bookedTime[i].end) {
                    reject('New event overlaps an existing event');
                    return;
                }
            }

            resolve();
        } catch (error) {
            reject(error);
        }
    });
}

async function insertEvent(event: CalendarEvent): Promise<{ event: CalendarEvent; googleEvent: EventType }> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            // insert event into google calendar
            let calendarUser = await getUser(event.user_id);
            if (calendarUser === undefined) {
                reject('User not found');
                return;
            }

            conn = await getConnection();

            const store = await getStoreByUser(calendarUser);

            const accessToken = await getAccessToken(store.owner_user_id);
            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            let activity = await getActivity(event.store_id, event.activity_id, true);
            let description = activity.name + '\n';
            description += 'E-Mail: ' + event.customer_email + '\n';
            description += '\n';
            description += '----\n';
            description += event.customer_message;

            const calendarID = await calendar_ids.getEventCalendarID(calendarUser);

            const requestBody = {
                summary: event.customer_name,
                description: description,
                start: {
                    dateTime: event.time_start.toISOString(),
                },
                end: {
                    dateTime: event.time_end.toISOString(),
                },
            };

            const calRes = await calendar.events.insert({
                calendarId: calendarID,
                requestBody,
            });

            if (calRes.data.id !== null) {
                await insertEventGoogleMariadb(calRes.data, calendarID, conn);
                event.calendar_event_id = calRes.data.id;
            }

            userLogger.info(event.user_id, 'Event inserted', JSON.stringify(event, null, 2));

            resolve({ event: event, googleEvent: calRes.data });
        } catch (error) {
            console.log('insertEvent error', error);

            reject(error);
        } finally {
            if (conn) conn.release();
            await clearRedisCache(event.store_id);
        }
    });
}

async function deleteEvent(event: CalendarEvent, ignoreIfNotExist?: boolean): Promise<void> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            // delete event from google calendar
            let calendarUser = await getUser(event.user_id);
            if (calendarUser === undefined) {
                reject('User not found');
                return;
            }

            conn = await getConnection();

            const store = await getStoreByUser(calendarUser);

            const calendarID = await calendar_ids.getEventCalendarID(calendarUser);

            if (event.calendar_event_id) await deleteEventGoogleMariadb(event.calendar_event_id, calendarID, conn);

            const accessToken = await getAccessToken(store.owner_user_id);
            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            await calendar.events.delete({
                calendarId: calendarID,
                eventId: event.calendar_event_id,
            });

            userLogger.info(event.user_id, 'Event deleted', JSON.stringify(event, null, 2));

            resolve();
        } catch (_error) {
            const error = _error as any;
            if (error.response && error.response.status === 410) {
                // 410 means event is already deleted
                resolve();
                return;
            }
            if (ignoreIfNotExist === true && error.response && error.response.status === 404) {
                resolve();
                return;
            }

            reject(error);
        } finally {
            if (conn) conn.release();
            await clearRedisCache(event.store_id);
        }
    });
}

export { checkIfTimeIsAvailable, insertEvent, deleteEvent };
