import { clientId, clientSecret } from '../config';

import GoogleTokensDatabase from '../database/user_google_tokens';

import { OAuth2Client } from 'google-auth-library';
import initCalendar from './initCalendar';
import { userLogger } from '../logger/logger';

function initOauthAccess(userId: string, refreshToken: string, accessToken: string, expiryDate: Date, googleUuid: string) {
    return new Promise(async (resolve, reject) => {
        // get access token

        const oauth2Client = new OAuth2Client({
            clientId,
            clientSecret,
        });

        oauth2Client.setCredentials({
            access_token: accessToken,
            refresh_token: refreshToken,
            //expiry_date: tokens.expiry_date.getTime(),
        });

        try {
            await GoogleTokensDatabase.initGoogleAccount(userId, accessToken, refreshToken, expiryDate, googleUuid);
            userLogger.info(userId, 'Google account initialized');

            try {
                await initCalendar(userId);

                await GoogleTokensDatabase.updateGoogleAccountStatus(userId, 'OK');
                resolve(accessToken);
            } catch (err) {
                console.log(err);
                if (err === '403') {
                    await GoogleTokensDatabase.updateGoogleAccountStatus(userId, 'NOPERM');
                } else {
                    await GoogleTokensDatabase.updateGoogleAccountStatus(userId, 'ERROR');
                }

                reject(err);
            }
        } catch (err) {
            console.log(err);

            await GoogleTokensDatabase.updateGoogleAccountStatus(userId, 'ERROR');
            reject(err);
        }
    });
}

async function getAccessToken(userId: string): Promise<string> {
    return new Promise(async (resolve, reject) => {
        GoogleTokensDatabase.getGoogleTokens(userId)
            .then((tokens) => {
                // check if token is expired
                const expiryDate = tokens.expiry_date;
                const now = new Date();

                //console.log('token expires in ' + (expiryDate.getTime() - now.getTime()) / 1000 / 60 + ' minutes');

                // refresh token if it expires in less than 1 minute
                if (expiryDate.getTime() - now.getTime() < 1 * 60 * 1000) {
                    //console.log('token expired, refreshing...');

                    const oauth2Client = new OAuth2Client({
                        clientId,
                        clientSecret,
                        //redirectUri,
                    });

                    oauth2Client.setCredentials({
                        access_token: tokens.access_token,
                        refresh_token: tokens.refresh_token,
                        expiry_date: tokens.expiry_date.getTime(),
                    });

                    oauth2Client
                        .refreshAccessToken()
                        .then((refreshedTokens: any) => {
                            const accessToken = refreshedTokens.credentials.access_token;
                            const expireDate = new Date(refreshedTokens.credentials.expiry_date);

                            // save to database
                            GoogleTokensDatabase.updateAccessToken(userId, accessToken, expireDate)
                                .then(() => {
                                    userLogger.info(userId, 'Google access token refreshed');
                                    resolve(accessToken);
                                })
                                .catch((err) => {
                                    console.log('updateAccessToken()', err);
                                    reject(err);
                                });
                        })
                        .catch((err: any) => {
                            if (err?.response?.data?.error === 'invalid_grant') {
                                GoogleTokensDatabase.unlinkGoogleAccount(userId);
                            } else {
                                userLogger.error(userId, 'Failed to refresh Google access token', JSON.stringify(err, null, 2));
                            }
                            reject(err);
                        });
                } else {
                    resolve(tokens.access_token);
                }
            })
            .catch((err) => {
                console.log('GetAccessToken', err);
                reject(err);
            });
    });
}

async function getOAuthClient(accessToken: string) {
    const auth = new OAuth2Client({
        clientId,
        clientSecret,
    });

    auth.setCredentials({ access_token: accessToken });

    return auth;
}

export { initOauthAccess, getAccessToken, getOAuthClient };
