import { google } from 'googleapis';
import { getAccessToken, getOAuthClient } from './oauthAccess';
import calendar_ids from '../database/calendar_ids';
import user_google_tokens from '../database/user_google_tokens';
import { getUser } from '../database/user';

import { User, UserID } from '../user/types';
import { StoreID, getAllStoreIDs, getStore } from '../database/store';
import { GLOBAL_calendar_min_earliest_booking_time } from '../config';

import dotenv from 'dotenv';
import { refreshOpenHours } from '../calendarSync/openHours';
import MyRedisKeys from '../redis/keys';
import redisClient from '../redis/init';
dotenv.config();

interface OpenHoursItem {
    start: Date;
    end: Date;
}

async function getOpenHours(storeID: string): Promise<OpenHoursItem[]> {
    return new Promise(async (resolve, reject) => {
        try {
            let cachedOpenHours = await redisClient.get(MyRedisKeys.calendarOpenHours(storeID));

            if (cachedOpenHours) {
                let openHours = JSON.parse(cachedOpenHours) as OpenHoursItem[];

                // change the date strings to Date objects
                openHours = openHours.map((openHour) => {
                    return {
                        start: new Date(openHour.start),
                        end: new Date(openHour.end),
                    };
                });
                resolve(openHours);
                return;
            }

            const openHours = await forceGetFreshOpenHours(storeID);

            resolve(openHours);
        } catch (error) {
            console.error(error);
            reject([]);
        }
    });
}

async function forceGetFreshOpenHours(storeID: string): Promise<OpenHoursItem[]> {
    return new Promise(async (resolve, reject) => {
        try {
            console.log('forceGetFreshOpenHours', storeID);
            const resp = await refreshOpenHours(storeID);
            resolve(resp || []);
        } catch (error) {
            console.error(error);
            reject([]);
        }
    });
}

function getClosedHours(openHours: OpenHoursItem[], earliestBookingTime: Date, endDate: Date) {
    let closedHours: OpenHoursItem[] = [];
    for (let i = 0; i < openHours.length; i++) {
        const openHour = openHours[i];

        // Skip the openHour if it's before the earliestBookingTime or after the endTime
        if (openHour.end < earliestBookingTime || openHour.start > endDate) {
            continue;
        }

        if (i == 0) {
            if (earliestBookingTime < openHour.start) {
                closedHours.push({ start: earliestBookingTime, end: openHour.start });
            }
        } else {
            const previousEnd = openHours[i - 1].end;
            // Only add to closedHours if the previous end time is within the booking window
            if (previousEnd >= earliestBookingTime && previousEnd <= endDate) {
                closedHours.push({ start: previousEnd, end: openHour.start });
            }
        }
    }

    return closedHours;
}

async function getStartOfDay(openHours: OpenHoursItem[], day: Date): Promise<Date> {
    return new Promise((resolve, reject) => {
        let startDate = null;
        let startOfDay = new Date(day);
        startOfDay.setHours(0, 0, 0, 0);
        const currentTime = new Date();

        for (let i = openHours.length - 1; i >= 0; i--) {
            const openHour = openHours[i];

            if (openHour.start <= startOfDay && openHour.end >= startOfDay) {
                startDate = startOfDay;
                break;
            }

            // find the first date in openHours that is after startOfDay
            if (openHour.start >= startOfDay && openHour.start <= (startDate || Infinity)) {
                startDate = openHour.start;
            } else {
                break;
            }
        }

        if (!startDate) {
            reject('The office is closed on this day');
            return;
        }

        // if startDate is before currentTime, set startDate to currentTime
        if (startDate < currentTime) {
            startDate = currentTime;
        }

        resolve(startDate);
    });
}

async function getStartOfDayOfUser(user: User, openHours: OpenHoursItem[], day: Date, doNotAddOffset?: boolean): Promise<Date> {
    const timeShiftOffset = doNotAddOffset ? 0 : GLOBAL_calendar_min_earliest_booking_time;
    let timeShift = (user.settings.calendar_min_earliest_booking_time + timeShiftOffset) * 60 * 1000;
    let endOfDay = await getEndOfDay(openHours, day);

    if (new Date().getTime() + timeShift > endOfDay.getTime()) {
        console.error('no earliestBookingTime found');
        throw new Error('The office is closed on this day');
    }

    const startOfDay = await getStartOfDay(openHours, day);

    if (new Date().getTime() + timeShift < startOfDay.getTime()) {
        timeShift = 0;
    }

    const earliestBookingTime = new Date(startOfDay.getTime() + timeShift);

    if (earliestBookingTime > endOfDay) {
        console.error('no earliestBookingTime found');
        throw new Error('The office is closed on this day');
    }
    return earliestBookingTime;
}

async function getEndOfDay(openHours: OpenHoursItem[], day: Date): Promise<Date> {
    return new Promise((resolve, reject) => {
        let endDate = null;
        let endOfDay = new Date(day);
        endOfDay.setHours(23, 59, 59, 999);
        const currentTime = new Date();

        for (let i = 0; i < openHours.length; i++) {
            const openHour = openHours[i];

            if (openHour.start <= endOfDay && openHour.end >= endOfDay) {
                endDate = endOfDay;
                break;
            }

            // find the last date in openHours that is before endOfDay
            if (openHour.end <= endOfDay && openHour.end >= (endDate || 0)) {
                if (openHour.end >= currentTime) {
                    endDate = openHour.end;
                }
            } else {
                break;
            }
        }

        if (!endDate) {
            console.error('no endDate found');
            reject('The office is closed on this day');
            return;
        }

        /*if (!endDate) {
            // find the last date in openHours
            for (let i = openHours.length - 1; i >= 0; i--) {
                const openHour = openHours[i];

                if (openHour.end >= (endDate || 0)) {
                    endDate = openHour.end;
                } else {
                    break;
                }
            }

            if (!endDate) {
                console.error('no endDate found');
                reject('The office is closed on this day');
                return;
            }

            // if currentTime is before endDate, set endDate to endOfDay
            if (currentTime < endDate) {
                endDate = endOfDay;
            } else {
                console.error('no endDate found');
                reject('The office is closed on this day');
                return;
            }
        }*/

        resolve(endDate);
    });
}

export { forceGetFreshOpenHours, getOpenHours, getEndOfDay, getStartOfDayOfUser, getClosedHours, OpenHoursItem };
