import { log } from 'console';
import { calendar_v3 } from 'googleapis';
import { PoolConnection } from 'mariadb';
import logger from '../logger/logger';

async function deleteEvent(eventId: string, calendarId: string, conn: PoolConnection) {
    return new Promise(async (resolve, reject) => {
        try {
            await conn.query('DELETE FROM `calendar_events` WHERE `event_id` = ? AND `calendar_id` = ?', [eventId, calendarId]);

            resolve(undefined);
        } catch (error) {
            console.log('deleteEvent error', error);
            logger.error('deleteEvent error', JSON.stringify(error, null, 2));
            reject(error);
        }
    });
}

// true if event is inserted, false if event is deleted
async function insertEvent(event: calendar_v3.Schema$Event, calendarId: string, conn: PoolConnection): Promise<boolean> {
    return new Promise(async (resolve, reject) => {
        try {
            if (event.status === 'cancelled') {
                try {
                    if (event.id) {
                        await deleteEvent(event.id, calendarId, conn);
                    }
                } catch (error) {}
                resolve(false);
                return;
            }

            let eventStart, eventEnd, eventUpdated, eventTransparency;

            if (event.start && event.start.date) {
                eventStart = new Date(event.start.date) || undefined;
                eventStart.setHours(0, 0, 0, 0);
            }
            if (event.start && event.start.dateTime) eventStart = new Date(event.start.dateTime) || undefined;

            if (event.end && event.end.date) {
                eventEnd = new Date(event.end.date) || undefined;
                eventEnd.setHours(0, 0, 0, 0);
            }
            if (event.end && event.end.dateTime) eventEnd = new Date(event.end.dateTime) || undefined;

            if (!event.transparency) eventTransparency = 'opaque';
            else eventTransparency = event.transparency;

            //if (event.updated) eventUpdated = new Date(event.updated);
            // use own updated time instead of google's
            eventUpdated = new Date();

            // insert event into database
            let query = 'INSERT INTO `calendar_events` (`event_id`, `calendar_id`, `start_time`, `end_time`, `status_val`, `updated_time`, `transparency`) VALUES (?, ?, ?, ?, ?, ?, ?)';
            let params = [event.id, calendarId, eventStart, eventEnd, event.status, eventUpdated, eventTransparency];

            let updates = [];
            if (eventStart !== null && eventStart !== undefined) {
                updates.push('`start_time` = ?');
                params.push(eventStart);
            }
            if (eventEnd !== null && eventEnd !== undefined) {
                updates.push('`end_time` = ?');
                params.push(eventEnd);
            }
            if (event.status !== null && event.status !== undefined) {
                updates.push('`status_val` = ?');
                params.push(event.status);
            }
            if (eventUpdated !== null && eventUpdated !== undefined) {
                updates.push('`updated_time` = ?');
                params.push(eventUpdated);
            }
            if (eventTransparency !== null && eventTransparency !== undefined) {
                updates.push('`transparency` = ?');
                params.push(eventTransparency);
            }

            if (updates.length > 0) {
                query += ' ON DUPLICATE KEY UPDATE ' + updates.join(', ');
            }

            await conn.query(query, params);

            resolve(true);
        } catch (error) {
            console.log('insertEvent error', error);
            reject(error);
        }
    });
}

export { insertEvent, deleteEvent };
