// External libraries
import { google } from 'googleapis';

// Calendar related imports
import { getAccessToken, getOAuthClient } from '../calendar/oauthAccess';

// Database related imports
import calendar_ids from '../database/calendar_ids';
import user_google_tokens from '../database/user_google_tokens';
import { getAllUsersIDsByStoreID, getUser } from '../database/user';
import { StoreID, getAllStoreIDs, getStore } from '../database/store';

// User related imports
import { User, UserID } from '../user/types';

// Event related imports
import { getAllEventsFromDatabase } from './getAllEvents';

// Redis related imports
import MyRedisKeys from '../redis/keys';
import redisClient from '../redis/init';

interface OpenHoursItem {
    start: Date;
    end: Date;
}

async function getHighestCalendarMaxFutureBookingDays(storeID: StoreID) {
    return new Promise<number>(async (resolve, reject) => {
        // get all users of this store
        const users = await getAllUsersIDsByStoreID(storeID);

        let highest_calendar_max_future_booking_days = 0;
        for (const user_id of users) {
            const user = await getUser(user_id);
            if (user.settings.calendar_max_future_booking_days > highest_calendar_max_future_booking_days) {
                highest_calendar_max_future_booking_days = user.settings.calendar_max_future_booking_days;
            }
        }

        resolve(highest_calendar_max_future_booking_days);
    });
}

async function refreshOpenHours(storeID: StoreID) {
    try {
        let openHoursList: OpenHoursItem[] = [];

        const store = await getStore(storeID);
        const user_id = store.owner_user_id;

        const highest_calendar_max_future_booking_days = await getHighestCalendarMaxFutureBookingDays(storeID);

        try {
            const calRes = await getAllEventsFromDatabase(
                await calendar_ids.getOpenHoursCalendar(user_id),
                new Date(),
                new Date(new Date().getTime() + highest_calendar_max_future_booking_days * 24 * 60 * 60 * 1000)
            );

            if (calRes) {
                const events = calRes;
                if (events.length) {
                    let eventList: OpenHoursItem[] = [];
                    for (let i = 0; i < events.length; i++) {
                        const event = events[i];
                        const start = event.start_time;
                        const end = event.end_time;

                        if (!start || !end) continue;

                        eventList.push({ start: new Date(start), end: new Date(end) });
                    }

                    let filteredEventList = [];
                    for (let i = 0; i < eventList.length; i++) {
                        const event = eventList[i];

                        let found = false;
                        for (let j = 0; j < filteredEventList.length; j++) {
                            const filteredEvent = filteredEventList[j];

                            if (event.start >= filteredEvent.start && event.start <= filteredEvent.end) {
                                if (event.end > filteredEvent.end) {
                                    filteredEvent.end = event.end;
                                }
                                found = true;
                                break;
                            } else if (event.end >= filteredEvent.start && event.end <= filteredEvent.end) {
                                if (event.start < filteredEvent.start) {
                                    filteredEvent.start = event.start;
                                }
                                found = true;
                                break;
                            } else if (event.start <= filteredEvent.start && event.end >= filteredEvent.end) {
                                filteredEvent.start = event.start;
                                filteredEvent.end = event.end;
                                found = true;
                                break;
                            }
                        }

                        if (!found) {
                            filteredEventList.push(event);
                        }
                    }

                    openHoursList = filteredEventList;

                    // save open hours to redis
                    const redisKey = MyRedisKeys.calendarOpenHours(storeID);
                    await redisClient.set(redisKey, JSON.stringify(openHoursList), { EX: 24 * 60 * 60 }); // 24 hours

                    return openHoursList;
                } else {
                    console.log('No upcoming events found.');
                }
            } else {
                console.log('No calendar response found.');
            }
        } catch (err) {
            console.log('The API returned an error: ' + err);
        }
    } catch (error) {
        console.log(error);
    }

    return null;
}

export { refreshOpenHours, getHighestCalendarMaxFutureBookingDays };
