// External libraries
import { calendar_v3, google } from 'googleapis';

// Database related imports
import { getConnection } from '../database/initDatabase';
import { getUser } from '../database/user';

// Calendar related imports
import { getAccessToken, getOAuthClient } from '../calendar/oauthAccess';

// Redis related imports
import redisClient from '../redis/init';
import MyRedisKeys from '../redis/keys';

// Config and utilities
import { GLOBAL_google_webhook_time_delay, GLOBAL_sync_calendar_every_minutes, googleSyncMaxDays, publicURL } from '../config';
import countUp from '../requestCounter';
import countGoogleRequestStatUp from '../requestCounter';

// Other internal modules
import { insertEvent } from './misc';
import { refreshOpenHours } from './openHours';
import generateRandomString from '../randomString';
import { PoolConnection } from 'mariadb';
import { User } from '../user/types';
import { forceGetFreshOpenHours } from '../calendar/openHoursList';
import clearRedisCache from '../redis/clearCache';
import fillCache from '../redis/fillCache';
import calendar_ids from '../database/calendar_ids';
import { storeLogger, userLogger } from '../logger/logger';
import GoogleTokensDatabase from '../database/user_google_tokens';

interface CalendarIDRow {
    user_id: string;
    calendar_id: string;
    type: string;
    worker_user_id?: string;
    sync_token?: string;
    next_sync?: Date;
    sync_token_expires_at?: Date;
    webhook_channel_id?: string;
    webhook_resource_id?: string;
    webhook_expires_at?: Date;
    isChanged?: boolean;
    isDisabled?: boolean;
}

async function refreshSyncToken() {
    let conn;
    try {
        conn = await getConnection();

        const currentTime = new Date(new Date().getTime() + 1000 * 60 * 30); // when it expires in 30 minutes

        const rows: CalendarIDRow[] = await conn.query('SELECT * FROM `calendar_ids` WHERE `is_disabled` IS NULL AND (`sync_token` IS NULL OR `sync_token_expires_at` < ?) ORDER BY RAND() LIMIT 1', [
            currentTime,
        ]);

        conn.end();
        //console.log('Found ' + rows.length + ' calendars without sync token');

        if (rows.length >= 1) {
            for (const row of rows) {
                try {
                    await syncCalendar(row, false);
                } catch (error) {
                    console.error(error);
                }
            }
        }
    } catch (error) {
        console.error(error);
    } finally {
        if (conn) {
            conn.end();
        }
    }
}

async function _updateSyncToken(_myCalendar: CalendarIDRow, user: User, calendarId: string, fullCalendarId: string, conn: PoolConnection, refreshLock: Function): Promise<CalendarIDRow> {
    return new Promise(async (resolve, reject) => {
        let myCalendar: CalendarIDRow = _myCalendar;

        try {
            const isPrimary = myCalendar.type === 'primary';
            if (isPrimary && user.settings.calendar_using_primary_calendar !== true) {
                resolve(myCalendar);
                return;
            }

            if (myCalendar.sync_token && myCalendar.sync_token_expires_at && myCalendar.sync_token_expires_at > new Date()) {
                resolve(myCalendar);
                return;
            }

            if (!myCalendar.sync_token || !myCalendar.sync_token_expires_at || myCalendar.sync_token_expires_at < new Date()) {
                await calendar_ids.clearCalendarDatabase(myCalendar.calendar_id);
                console.log('Cleared calendar database for calendar ' + myCalendar.calendar_id);
                userLogger.info(user.user_id, '_updateSyncToken()', 'Cleared calendar database for calendar ' + myCalendar.calendar_id);
            }

            const next_sync = new Date(new Date().getTime() + 1000 * 60 * 15); // first time sync again in 15 minutes

            const accessToken = await getAccessToken(user.user_id);
            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            // create calendar sync token for row.calendar_id
            const timeMax = new Date(new Date().getTime() + googleSyncMaxDays * 24 * 60 * 60 * 1000);
            const expireTime = new Date(new Date().getTime() + 14 * 24 * 60 * 60 * 1000); // make new whole sync every 14 days
            let getListSettings: calendar_v3.Params$Resource$Events$List = {
                calendarId: calendarId,
                timeMin: new Date().toISOString(),
                timeMax: timeMax.toISOString(),
                maxResults: 2500, // warning: this is a hard limit
                singleEvents: true,
                showDeleted: true,
            };

            userLogger.info(user.user_id, '_updateSyncToken()', "running full sync for calendar '" + fullCalendarId + "'");

            // get calendar events
            countGoogleRequestStatUp('refreshSyncToken');
            let events = await calendar.events.list(getListSettings);

            const totalEvents = [events];

            let pageCounter = 0;

            // get all events
            while (events.data.nextPageToken) {
                pageCounter++;
                // refresh lock
                await refreshLock();

                // wait random time between 0 and 10 seconds
                conn.end();
                await new Promise((resolve) => setTimeout(resolve, Math.random() * 10000));
                conn = await getConnection();

                getListSettings.pageToken = events.data.nextPageToken;
                countGoogleRequestStatUp('refreshSyncTokenPage');
                events = await calendar.events.list(getListSettings);

                totalEvents.push(events);
            }

            // get sync token
            const newSyncToken = events.data.nextSyncToken;

            userLogger.info(user.user_id, '_updateSyncToken()', "calendar '" + fullCalendarId + "' took " + (pageCounter + 1) + ' pages to sync');

            if (newSyncToken) {
                let insertedEvents = 0;
                let deletedEvents = 0;

                for (const _events of totalEvents) {
                    if (!_events.data.items) continue;

                    for (const event of _events.data.items) {
                        // insert event into database

                        const isInsered = await insertEvent(event, fullCalendarId, conn);
                        if (isInsered) {
                            insertedEvents++;
                        } else {
                            // deleted
                            deletedEvents++;
                        }
                    }
                }

                userLogger.info(
                    user.user_id,
                    '_updateSyncToken()',
                    "calendar '" + fullCalendarId + "' took " + (pageCounter + 1) + ' pages to sync and inserted ' + insertedEvents + ' events and deleted ' + deletedEvents + ' events'
                );

                // update sync token
                await conn.query('UPDATE `calendar_ids` SET `sync_token` = ?, `sync_token_expires_at` = ?, `next_sync` = ? WHERE `calendar_id` = ?', [
                    newSyncToken,
                    expireTime /*also min. sync_token_expires_at*/,
                    next_sync,
                    fullCalendarId,
                ]);

                myCalendar.sync_token = newSyncToken;
                myCalendar.sync_token_expires_at = expireTime;
                myCalendar.next_sync = next_sync;
            } else {
                console.error('newSyncToken is null');
                myCalendar.sync_token = undefined;
                myCalendar.sync_token_expires_at = undefined;
                myCalendar.next_sync = undefined;
            }

            myCalendar.isChanged = true;

            resolve(myCalendar);
        } catch (error) {
            if ((error as any).code === 403 || (error as any).code === 404) {
                console.error('403 error, setting is_disabled to 1');
                userLogger.error(user.user_id, '_updateSyncToken()', '403 error, setting is_disabled to 1');
                await conn.query('UPDATE `calendar_ids` SET `is_disabled` = 1 WHERE `calendar_id` = ?', [fullCalendarId]);
                myCalendar.isDisabled = true;
                myCalendar.isChanged = true;

                resolve(myCalendar);
            } else {
                reject(error);
            }
        }
    });
}

async function _updateEvents(_myCalendar: CalendarIDRow, user: User, calendarId: string, fullCalendarId: string, conn: PoolConnection, refreshLock: Function): Promise<CalendarIDRow> {
    return new Promise(async (resolve, reject) => {
        try {
            let myCalendar: CalendarIDRow = _myCalendar;
            const isPrimary = myCalendar.type === 'primary';
            if (isPrimary && user.settings.calendar_using_primary_calendar !== true) {
                resolve(myCalendar);
                return;
            }

            if (!myCalendar.sync_token || !myCalendar.sync_token_expires_at || myCalendar.sync_token_expires_at < new Date()) {
                resolve(myCalendar);
                return;
            }

            if (myCalendar.next_sync && myCalendar.next_sync > new Date()) {
                resolve(myCalendar);
                return;
            }

            const accessToken = await getAccessToken(user.user_id);
            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            const getListSettings: calendar_v3.Params$Resource$Events$List = {
                calendarId: calendarId,
                syncToken: myCalendar.sync_token,
                maxResults: 2500, // warning: this is a hard limit
                singleEvents: true,
                showDeleted: true,
            };

            // get calendar events
            countGoogleRequestStatUp('updateEvents');
            let events = await calendar.events.list({
                ...getListSettings,
            });

            const totalEvents = [events];

            let pageCounter = 0;

            // get all events
            while (events.data.nextPageToken) {
                pageCounter++;

                // refresh lock
                await refreshLock();

                // wait random time between 0 and 10 seconds
                conn.end();
                await new Promise((resolve) => setTimeout(resolve, Math.random() * 10000));
                conn = await getConnection();

                countGoogleRequestStatUp('updateEventsPage');
                events = await calendar.events.list({
                    pageToken: events.data.nextPageToken,
                    ...getListSettings,
                });

                totalEvents.push(events);
            }

            // update sync token
            const syncToken = events.data.nextSyncToken;
            const next_sync = new Date(new Date().getTime() + 1000 * 60 * GLOBAL_sync_calendar_every_minutes);

            if (syncToken && syncToken !== myCalendar.sync_token) {
                // update sync token in database
                await conn.query('UPDATE `calendar_ids` SET `sync_token` = ? WHERE `calendar_id` = ?', [syncToken, fullCalendarId]);

                console.log('Updated sync token for calendar ' + fullCalendarId, 'from', myCalendar.sync_token, 'to', syncToken);
                myCalendar.sync_token = syncToken;
            }

            let insertedEvents = 0;
            let deletedEvents = 0;

            // update events
            for (const _events of totalEvents) {
                if (!_events.data.items) continue;

                for (const event of _events.data.items) {
                    // insert event into database
                    console.log('inserting event', event);
                    const isInsered = await insertEvent(event, fullCalendarId, conn);

                    if (isInsered) {
                        insertedEvents++;
                    } else {
                        // deleted
                        deletedEvents++;
                    }
                }
            }

            userLogger.info(
                user.user_id,
                '_updateEvents()',
                "calendar '" + fullCalendarId + "' took " + (pageCounter + 1) + ' pages to sync and inserted ' + insertedEvents + ' events and deleted ' + deletedEvents + ' events'
            );

            // update next_sync
            await conn.query('UPDATE `calendar_ids` SET `next_sync` = ? WHERE `calendar_id` = ?', [next_sync, fullCalendarId]);
            console.log('Updated next_sync for calendar ' + fullCalendarId, 'to', next_sync);

            myCalendar.next_sync = next_sync;
            myCalendar.isChanged = true;

            resolve(myCalendar);
        } catch (error) {
            reject(error);
        }
    });
}

async function _subscribeToWebhook(_myCalendar: CalendarIDRow, user: User, calendarId: string, fullCalendarId: string, conn: PoolConnection, refreshLock: Function): Promise<CalendarIDRow> {
    return new Promise(async (resolve, reject) => {
        try {
            let myCalendar: CalendarIDRow = _myCalendar;
            const isPrimary = myCalendar.type === 'primary';
            if (isPrimary && user.settings.calendar_using_primary_calendar !== true) {
                resolve(myCalendar);
                return;
            }

            if (!myCalendar.sync_token || !myCalendar.sync_token_expires_at || myCalendar.sync_token_expires_at < new Date()) {
                resolve(myCalendar);
                return;
            }

            if (myCalendar.webhook_channel_id && myCalendar.webhook_expires_at && myCalendar.webhook_expires_at > new Date()) {
                resolve(myCalendar);
                return;
            }

            const accessToken = await getAccessToken(user.user_id);
            const calendar = google.calendar({ version: 'v3', auth: await getOAuthClient(accessToken) });

            const webhookId = await generateRandomString(32);

            const webhook = await calendar.events.watch({
                calendarId: calendarId,
                requestBody: {
                    id: webhookId,
                    type: 'web_hook',
                    address: publicURL + '/api/v1/notification/' + encodeURIComponent(fullCalendarId),
                    token: process.env.GOOGLE_WEBHOOK_SECRET,
                    //expiration: 86400, // 24 hours
                },
                singleEvents: true,
                showDeleted: true,
                timeMin: new Date().toISOString(),
                timeMax: myCalendar.sync_token_expires_at.toISOString(),
            });

            console.log('webhook', webhook);
            console.log('webhook', webhook.data);

            const webhookResourceId = webhook.data.resourceId;
            let webhookExpiration;
            if (Number(webhook.data.expiration) > 0) {
                webhookExpiration = new Date(Number(webhook.data.expiration));
            } else {
                webhookExpiration = new Date(new Date().getTime() + 86400 * 1000); // 24 hours
            }

            // update webhook channel id and expiration
            await conn.query('UPDATE `calendar_ids` SET `webhook_channel_id` = ?, `webhook_resource_id` = ?, `webhook_expires_at` = ? WHERE `calendar_id` = ?', [
                webhookId,
                webhookResourceId,
                webhookExpiration,
                fullCalendarId,
            ]);

            myCalendar.webhook_channel_id = webhookId;
            myCalendar.webhook_resource_id = webhookResourceId || undefined;
            myCalendar.webhook_expires_at = webhookExpiration;

            userLogger.info(user.user_id, '_subscribeToWebhook()', 'Subscribed to webhook for calendar ' + fullCalendarId, 'with webhook id', webhookId);

            resolve(myCalendar);
        } catch (error) {
            reject(error);
        }
    });
}

async function _checkifAccessTokenIsValid(_myCalendar: CalendarIDRow, user: User, calendarId: string, fullCalendarId: string, conn: PoolConnection): Promise<CalendarIDRow> {
    return new Promise(async (resolve, reject) => {
        let myCalendar: CalendarIDRow = _myCalendar;

        try {
            // this can be used because when user.settings.calendar_using_primary_calendar is changed, the backend won't recognize it until the next sync

            /*if (myCalendar.type === 'primary' && user.settings.calendar_using_primary_calendar !== true) {
                throw new Error('Primary calendar is not used');
            }*/

            await getAccessToken(user.user_id);
            resolve(myCalendar);
        } catch (error) {
            try {
                // set  `is_disabled` to 1 in `calendar_ids` table
                await conn.query('UPDATE `calendar_ids` SET `is_disabled` = 1 WHERE `calendar_id` = ?', [fullCalendarId]);
                myCalendar.isDisabled = true;
                myCalendar.isChanged = true;

                resolve(myCalendar);
            } catch (error) {
                reject(error);
            }
        }
    });
}

async function enableCalendarSync(calendarId: string) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            const rows = await conn.query('UPDATE `calendar_ids` SET `is_disabled` = NULL WHERE `calendar_id` = ?', [calendarId]);

            resolve(rows);
        } catch (error) {
            reject(error);
        } finally {
            if (conn) {
                conn.end();
            }
        }
    });
}

async function syncCalendar(_myCalendar: CalendarIDRow, waitForLock: boolean = true) {
    return new Promise(async (resolve, reject) => {
        let myCalendar = { ..._myCalendar };

        const redisLockKey = MyRedisKeys.calendarSyncLock(myCalendar.calendar_id);
        const redisLockKeyTTL = 2 * 60; // 2 minutes

        let isRedisLocked = null;

        function lock() {
            return redisClient.set(redisLockKey, 'true', { EX: redisLockKeyTTL, NX: true });
        }

        function refreshLock() {
            return redisClient.expire(redisLockKey, redisLockKeyTTL);
        }

        try {
            isRedisLocked = await lock();
        } catch (error) {
            console.error(error);
            reject(error);
            return;
        }

        // if lock is not set
        if (isRedisLocked === null) {
            if (waitForLock) {
                while (isRedisLocked === null) {
                    // Wait for a short period before trying again
                    await new Promise((resolve) => setTimeout(resolve, 1000)); // wait for 1 second

                    // Try to acquire the lock again
                    isRedisLocked = await lock();
                }
            } else {
                reject('Lock is already set');
                return;
            }
        }

        let conn;

        try {
            conn = await getConnection();

            let calendarId = myCalendar.calendar_id;
            let fullCalendarId = myCalendar.calendar_id;
            let userId = myCalendar.user_id;

            const user = await getUser(myCalendar.user_id);

            const isPrimary = myCalendar.type === 'primary';
            if (isPrimary) {
                calendarId = fullCalendarId.split('#')[0];
                userId = fullCalendarId.split('#')[1];
            }

            try {
                myCalendar = await _checkifAccessTokenIsValid(myCalendar, user, calendarId, fullCalendarId, conn);

                if (myCalendar.isDisabled !== true) {
                    myCalendar = await _updateSyncToken(myCalendar, user, calendarId, fullCalendarId, conn, refreshLock);
                    myCalendar = await _updateEvents(myCalendar, user, calendarId, fullCalendarId, conn, refreshLock);
                    myCalendar = await _subscribeToWebhook(myCalendar, user, calendarId, fullCalendarId, conn, refreshLock);
                }
            } catch (error) {
                console.error(error);

                if ((error as any).code === 410) {
                    // 410 is the code for the sync token being invalid
                    myCalendar.sync_token = undefined;
                    myCalendar.sync_token_expires_at = undefined;
                    myCalendar.next_sync = undefined;
                    myCalendar.isChanged = true;

                    await conn.query('UPDATE `calendar_ids` SET `sync_token` = NULL, `sync_token_expires_at` = NULL, `next_sync` = NULL WHERE `calendar_id` = ?', [fullCalendarId]);

                    userLogger.error(user.user_id, 'syncCalendar() error 410', 'Sync token is invalid');
                } else if ((error as any).code === 401) {
                    // 401 is the code for the access token being invalid
                    await conn.query('UPDATE `calendar_ids` SET `is_disabled` = 1 WHERE `calendar_id` = ?', [fullCalendarId]);
                    myCalendar.isDisabled = true;
                    myCalendar.isChanged = true;

                    userLogger.error(user.user_id, 'syncCalendar() error 401', 'Access token is invalid');

                    // set `next_sync` = NULL for all calendars of this user
                    await conn.query('UPDATE `calendar_ids` SET `next_sync` = NULL WHERE `user_id` = ?', [user.user_id]);

                    if (user.isOwner) {
                        if (myCalendar.type === 'openHours') {
                            await GoogleTokensDatabase.updateGoogleAccountStatus(user.user_id, 'NOPERM');
                            storeLogger.error(user.store_id, 'syncCalendar() error 401', 'Access token is invalid', "set store status to 'NOPERM'");
                        }
                    } else {
                        await GoogleTokensDatabase.updateGoogleAccountStatus(user.user_id, 'NOPERM');
                        storeLogger.error(user.store_id, 'syncCalendar() error 401', 'Access token is invalid', "set store status to 'NOPERM'");
                    }
                } else {
                    let errorCode = (error as any).code !== undefined && (error as any).code !== null ? (error as any).code : 'unknown';
                    userLogger.error(user.user_id, 'syncCalendar() error', errorCode, (error as Error).toString());
                }
            }

            if (myCalendar.isChanged === true) {
                // clear whole redis cache of this store
                await clearRedisCache(user.store_id);

                // update open hours
                if (myCalendar.type === 'openHours') {
                    await forceGetFreshOpenHours(user.store_id);
                }

                setTimeout(() => {
                    fillCache(user.store_id);
                }, 500);
            }

            resolve(myCalendar);
        } catch (error) {
            userLogger.error(myCalendar.user_id, 'syncCalendar()', 'Error', JSON.stringify(error, null, 2));
            console.log('syncCalendar()');
            console.error(error);

            reject(error);
        } finally {
            if (conn) {
                conn.end();
            }

            // remove lock
            await redisClient.del(redisLockKey);
        }
    });
}

async function scheduleCalendarSync(fullcalendarId: string, channel_id: string) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            let rows = await conn.query('SELECT * FROM `calendar_ids` WHERE `calendar_id` = ?', [fullcalendarId]);

            if (rows.length === 0) {
                reject('Calendar not found');
                return;
            }

            const row = rows[0];

            if (row.webhook_channel_id !== channel_id) {
                resolve(false);
                return;
            }

            const next_sync = new Date(new Date().getTime() + 1000 * GLOBAL_google_webhook_time_delay);

            await conn.query('UPDATE `calendar_ids` SET `next_sync` = ? WHERE `calendar_id` = ?', [next_sync, fullcalendarId]);

            resolve(true);
        } catch (error) {
            reject(error);
        } finally {
            if (conn) {
                conn.end();
            }
        }
    });
}

export default syncCalendar;
export { refreshSyncToken, scheduleCalendarSync, enableCalendarSync };
