import { User } from '../user/types';
import { getConnection } from './initDatabase';
import { getStoreByUser } from './store';

type CalendarType = 'openHours' | 'myEvents' | 'workerEvents' | 'primary';

async function addCalendar(user_id: string, calendar_id: string, type: CalendarType, worker_user_id?: string) {
    return new Promise(async (resolve, reject) => {
        const _worker_user_id = worker_user_id ? worker_user_id : null;
        if (type == 'workerEvents' && _worker_user_id == null) {
            reject('worker_user_id must be set for type workerEvents');
            return;
        }

        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query(
                `
                INSERT INTO calendar_ids (user_id, calendar_id, type, worker_user_id) 
                VALUES (?, ?, ?, ?) 
                ON DUPLICATE KEY UPDATE 
                    user_id = VALUES(user_id), 
                    calendar_id = VALUES(calendar_id), 
                    type = VALUES(type), 
                    worker_user_id = VALUES(worker_user_id),
                    sync_token = NULL,
                    is_disabled = NULL
            `,
                [user_id, calendar_id, type, _worker_user_id]
            );
            console.log(rows);

            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

type Calendar = {
    user_id: string;
    calendar_id: string;
    type: CalendarType;
    worker_user_id?: string; // only for workerEvents
};

async function getCalendars(user_id: string): Promise<Calendar[]> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT * FROM calendar_ids WHERE user_id=?', [user_id]);

            let calendars: Calendar[] = [];

            for (let i = 0; i < rows.length; i++) {
                const calendar: Calendar = rows[i];

                if (calendar.user_id != user_id) {
                    console.error('user_id does not match');
                    continue;
                }

                const type = calendar.type;

                if (type === 'workerEvents' && calendar.worker_user_id === null) {
                    console.error('worker_user_id must be set for type workerEvents. calendar.calendar_id:', calendar.calendar_id);
                    continue;
                }

                if (type != 'workerEvents' && type != 'openHours' && type != 'myEvents' && type != 'primary') {
                    console.error('invalid type "' + type + '" calendar.calendar_id:', calendar.calendar_id);
                    continue;
                }

                calendars.push(calendar);
            }

            resolve(calendars);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getPrimaryCalendar(user_id: string): Promise<Calendar> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT calendar_id FROM calendar_ids WHERE user_id=? AND type="primary" AND is_disabled IS NULL', [user_id]);

            if (rows.length >= 1) {
                let calendar: Calendar = rows[0];

                //calendar.calendar_id = calendar.calendar_id.split('#')[0];

                resolve(calendar);
            } else {
                reject('no calendar found');
            }
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function deleteCalendar(user: User, calendar_id: string) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            let rows;

            if (user.isOwner === true) {
                await clearCalendarDatabase(calendar_id);
                rows = await conn.query('DELETE FROM calendar_ids WHERE user_id=? AND calendar_id=?', [user.user_id, calendar_id]);
            } else {
                const store = await getStoreByUser(user);
                const owner_user_id = store.owner_user_id;
                await clearCalendarDatabase(calendar_id);
                rows = await conn.query('DELETE FROM calendar_ids WHERE user_id=? AND calendar_id=? AND worker_user_id=?', [owner_user_id, calendar_id, user.user_id]);
            }

            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function clearCalendarDatabase(calendar_id: string) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            await conn.query('DELETE FROM calendar_events WHERE calendar_id=?', [calendar_id]);
            resolve(undefined);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getOpenHoursCalendar(user_id: string): Promise<string> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT calendar_id FROM calendar_ids WHERE user_id=? AND type="openHours"', [user_id]);

            if (rows.length >= 1) {
                resolve(rows[0].calendar_id);
            } else {
                reject('no calendar found');
            }
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getEventCalendarID(user: User): Promise<string> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            let rows;
            if (user.isOwner === true) rows = await conn.query('SELECT calendar_id FROM calendar_ids WHERE user_id=? AND type="myEvents" AND is_disabled IS NULL', [user.user_id]);
            else {
                let store = await getStoreByUser(user);
                if (!store) {
                    reject('no store found');
                    return;
                }

                let owner_user_id = store.owner_user_id;

                rows = await conn.query('SELECT calendar_id FROM calendar_ids WHERE user_id=? AND type="workerEvents" AND worker_user_id=? AND is_disabled IS NULL', [owner_user_id, user.user_id]);
            }

            if (rows.length >= 1) {
                resolve(rows[0].calendar_id);
            } else {
                reject('no calendar found');
            }
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function checkIfCalendarExists(calendarId: string): Promise<boolean> {
    return new Promise<boolean>(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            const rows = await conn.query('SELECT * FROM `calendar_ids` WHERE `calendar_id` = ? AND `is_disabled` IS NULL', [calendarId]);

            if (rows.length >= 1) {
                resolve(true);
            } else {
                resolve(false);
            }
        } catch (error) {
            console.log('checkIfCalendarExists error', error);
            reject(error);
        } finally {
            if (conn) {
                conn.end();
            }
        }
    });
}

const calendar_ids = {
    addCalendar,
    getCalendars,
    deleteCalendar,
    getOpenHoursCalendar,
    getEventCalendarID,
    getPrimaryCalendar,
    checkIfCalendarExists,
    clearCalendarDatabase,
};

export default calendar_ids;
export type { CalendarType };
