import { User } from '../user/types';
import { getConnection } from './initDatabase';
import { Store } from './store';

import { GLOBAL_calendar_min_earliest_booking_time, domain } from '../config';

import { deleteEvent, insertEvent } from '../calendar/insertEvent';
import calendar_ids from './calendar_ids';
import { getUser } from './user';
import sendAppointmentCanceledMail from '../mail/verifyFail';
import { sendMailIsVerifiedMail, sendNewAppointmentForEmployeeMail } from '../mail/verify';
import generateRandomString from '../randomString';
import clearRedisCache from '../redis/clearCache';
import { getActivity } from './services';

type CustomerID = string;

interface CalendarEvent {
    store_id: string;
    user_id: string;
    activity_id: string;
    calendar_event_id?: string;
    time_start: Date;
    time_end: Date;
    customer_email: string;
    customer_name: string;
    customer_message?: string;
    customer_id: CustomerID;
    customer_verified_time?: Date;
    send_next_notification_at?: Date;
    created_at: Date;
}

interface PublicCalendarEvent {
    username: string;
    activity_name: string;
    time_start: Date;
    time_end: Date;
    isVerified: boolean;
}

async function createEvent(
    store: Store,
    user: User,
    activity_id: string,
    time_start: Date,
    time_end: Date,
    customer_email: string,
    customer_name: string,
    customer_message?: string
): Promise<CalendarEvent> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            const created_at = new Date();

            // create a uuid for the customer email
            const customer_uuid = await generateRandomString(128);

            let calendarEvent: CalendarEvent = {
                store_id: store.store_id,
                user_id: user.user_id,
                activity_id: activity_id,
                time_start: time_start,
                time_end: time_end,
                customer_email: customer_email,
                customer_name: customer_name,
                customer_message: customer_message,
                customer_id: customer_uuid,

                created_at: created_at,
            };

            const insertedEvent = await insertEvent(calendarEvent);

            calendarEvent = insertedEvent.event;

            const rows = await conn.query(
                `INSERT INTO calendar_customer 
                    (store_id, name, email, message, customer_id, calendar_event_id, user_id, activity_id, time_start, time_end, created_at) 
                VALUES 
                    (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)`,
                [store.store_id, customer_name, customer_email, customer_message, customer_uuid, calendarEvent.calendar_event_id, user.user_id, activity_id, time_start, time_end, created_at]
            );

            resolve(calendarEvent);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getEvent(customer_id: CustomerID): Promise<CalendarEvent | undefined> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            const rows = await conn.query(`SELECT * FROM calendar_customer WHERE customer_id=?`, [customer_id]);

            if (rows.length < 1) {
                resolve(undefined);
                return;
            }

            const row = rows[0];

            const calendarEvent: CalendarEvent = {
                store_id: row.store_id,
                user_id: row.user_id,
                activity_id: row.activity_id,
                time_start: row.time_start,
                time_end: row.time_end,
                customer_email: row.email,
                calendar_event_id: row.calendar_event_id,
                customer_name: row.name,
                customer_message: row.message,
                customer_id: row.customer_id,
                customer_verified_time: row.verified_time,
                created_at: row.created_at,
            };

            resolve(calendarEvent);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getPublicEvent(customer_id: CustomerID): Promise<PublicCalendarEvent | undefined> {
    return new Promise(async (resolve, reject) => {
        let calendarEvent = await getEvent(customer_id);

        if (calendarEvent === undefined) {
            resolve(undefined);
            return;
        }

        let username = 'unknown';
        let activity_name = 'unknown';

        try {
            let user = await getUser(calendarEvent.user_id);
            username = user.username;
        } catch (error) {}

        try {
            let activity = await getActivity(calendarEvent.store_id, calendarEvent.activity_id, false);
            activity_name = activity.name;
        } catch (error) {}

        let publicCalendarEvent: PublicCalendarEvent = {
            username: username,
            activity_name: activity_name,
            time_start: calendarEvent.time_start,
            time_end: calendarEvent.time_end,
            isVerified: calendarEvent.customer_verified_time !== undefined && calendarEvent.customer_verified_time !== null,
        };

        resolve(publicCalendarEvent);
    });
}

async function verifyEvent(customer_id: CustomerID): Promise<CalendarEvent | undefined> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            const verified_time = new Date();
            let isNewVerify = false;

            // also select the event
            const calendarEventRows = await conn.query(`SELECT * FROM calendar_customer WHERE customer_id=?`, [customer_id]);

            if (calendarEventRows.length != 1) {
                resolve(undefined);
                return;
            }
            const calendarEventRow = calendarEventRows[0];

            if (calendarEventRow.verified_time === null || calendarEventRow.verified_time === undefined) {
                const rows = await conn.query(`UPDATE calendar_customer SET verified_time=? WHERE customer_id=?`, [verified_time, customer_id]);
                isNewVerify = true;
            }

            const calendarEvent: CalendarEvent = {
                store_id: calendarEventRow.store_id,
                user_id: calendarEventRow.user_id,
                activity_id: calendarEventRow.activity_id,
                time_start: calendarEventRow.time_start,
                time_end: calendarEventRow.time_end,
                customer_email: calendarEventRow.email,
                calendar_event_id: calendarEventRow.calendar_event_id,
                customer_name: calendarEventRow.name,
                customer_message: calendarEventRow.message,
                customer_id: calendarEventRow.customer_id,
                customer_verified_time: calendarEventRow.verified_time,
                created_at: calendarEventRow.created_at,
            };

            resolve(calendarEvent);

            try {
                if (isNewVerify === true) {
                    await sendMailIsVerifiedMail(calendarEvent);
                }
            } catch (error) {
                console.log('error sending sendMailIsVerifiedMail');
                console.error(error);
            }
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function cancelEvent(customer_id: CustomerID): Promise<CalendarEvent | undefined> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            // also select the event
            const calendarEventRows = await conn.query(`SELECT * FROM calendar_customer WHERE customer_id=?`, [customer_id]);

            console.log('calendarEventRows', calendarEventRows);

            if (calendarEventRows.length != 1) {
                resolve(undefined);
                return;
            }

            const calendarEventRow = calendarEventRows[0];

            const calendarEvent: CalendarEvent = {
                store_id: calendarEventRow.store_id,
                user_id: calendarEventRow.user_id,
                activity_id: calendarEventRow.activity_id,
                time_start: calendarEventRow.time_start,
                time_end: calendarEventRow.time_end,
                customer_email: calendarEventRow.email,
                calendar_event_id: calendarEventRow.calendar_event_id,
                customer_name: calendarEventRow.name,
                customer_message: calendarEventRow.message,
                customer_id: calendarEventRow.customer_id,
                customer_verified_time: calendarEventRow.verified_time,
                created_at: calendarEventRow.created_at,
            };

            await deleteEvent(calendarEvent, true);

            await sendAppointmentCanceledMail(calendarEvent);
            await conn.query(`DELETE FROM calendar_customer WHERE customer_id=?`, [customer_id]);

            resolve(calendarEvent);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

function deleteAllThatIsUnverified(): Promise<void> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            const dateTimeout = new Date().getTime() - GLOBAL_calendar_min_earliest_booking_time * 60 * 1000;

            const rows = await conn.query(`SELECT * FROM calendar_customer WHERE verified_time IS NULL AND created_at < ?`, [new Date(dateTimeout)]);

            for (let i = 0; i < rows.length; i++) {
                const row = rows[i];

                const calendarEvent: CalendarEvent = {
                    store_id: row.store_id,
                    user_id: row.user_id,
                    activity_id: row.activity_id,
                    time_start: row.time_start,
                    time_end: row.time_end,
                    customer_email: row.email,
                    calendar_event_id: row.calendar_event_id,
                    customer_name: row.name,
                    customer_message: row.message,
                    customer_id: row.customer_id,
                    customer_verified_time: row.verified_time,
                    created_at: row.created_at,
                };

                try {
                    if (calendarEvent.calendar_event_id !== undefined) {
                        await deleteEvent(calendarEvent, true);
                    }

                    await sendAppointmentCanceledMail(calendarEvent);
                    await conn.query(`DELETE FROM calendar_customer WHERE customer_id=?`, [row.customer_id]);
                } catch (error) {
                    console.log('error deleting unverified event');
                    console.error(error);
                }
            }

            resolve();
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

function generateEMailVerificationLink(customer_id: CustomerID): string {
    return process.env.REACT_STATUS_URL + '/?id=' + encodeURIComponent(customer_id);
}

function generateEMailCancelLink(customer_id: CustomerID): string {
    return process.env.REACT_STATUS_URL + '/?id=' + encodeURIComponent(customer_id) + '&cancel=true';
}

function generateEMailCancelNotificationLink(customer_id: CustomerID): string {
    return process.env.REACT_STATUS_URL + '/?id=' + encodeURIComponent(customer_id) + '&notify=false';
}

setInterval(deleteAllThatIsUnverified, 1000 * 60); // every minute
deleteAllThatIsUnverified();

const MyEvents = {
    createEvent: createEvent,
    verifyEvent: verifyEvent,
    cancelEvent: cancelEvent,
    generateEMailVerificationLink: generateEMailVerificationLink,
    generateEMailCancelLink: generateEMailCancelLink,
    generateEMailCancelNotificationLink: generateEMailCancelNotificationLink,
    getEvent: getEvent,
    getPublicEvent: getPublicEvent,
};

export { CalendarEvent };
export default MyEvents;

/*
CREATE TABLE `calendar_customer` (
  `store_id` varchar(255) NOT NULL,
  `name` varchar(64) NOT NULL,
  `email` varchar(255) NOT NULL,
  `message` varchar(1000) DEFAULT NULL,
  `customer_id` varchar(64) NOT NULL PRIMARY KEY,
  `calendar_event_id` varchar(255) NOT NULL,
  `user_id` varchar(255) NOT NULL,
  `activity_id` varchar(255) NOT NULL,
  `time_start` datetime NOT NULL,
  `time_end` datetime NOT NULL,
  `verified_time` datetime DEFAULT NULL,
  `created_at` datetime NOT NULL
)
*/
