import { defaultUserSettings } from '../config';
import { User, UserID } from '../user/types';
import { getConnection } from './initDatabase';
import { StoreID } from './store';
import { getAllUsersIDsAndNamesByStoreID, getAllUsersIDsByStoreID } from './user';

type ServicesTypeId = string;
type ActivityTypeId = string;

interface ServicesType {
    id: ServicesTypeId;
    name: string;
    activities: ActivityType[];
}

type ActivityUsersType = UserID[];

interface ActivityType {
    id: ActivityTypeId;
    name: string;
    description: string;
    price: number;
    duration: number;
    users: ActivityUsersType;
}

// displayMinimal: if all employees are assinged to the activity, the users object will be empty
async function getUserIDsByActivityID(store_id: StoreID, activity_id: ActivityTypeId, displayMinimal: boolean): Promise<ActivityUsersType> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();

            let userIDs: ActivityUsersType = [];

            let userRows: { user_id: UserID }[] = await conn.query('SELECT user_id FROM store_service_activity_users WHERE activity_id=?', [activity_id]);

            let allUserIDs: ActivityUsersType = await getAllUsersIDsByStoreID(store_id);

            if (displayMinimal === true) {
                if (!userRows || userRows.length === 0) userIDs = [];
                else userIDs = userRows.map((row) => row.user_id);
            } else {
                if (!userRows || userRows.length === 0) userIDs = allUserIDs;
                else userIDs = userRows.map((row) => row.user_id);
            }

            resolve(userIDs);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

// displayMinimal: if all employees are assinged to the activity, the users object will be empty
async function getServices(store_id: StoreID, displayMinimal: boolean): Promise<ServicesType[]> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT * FROM store_services WHERE store_id=?', [store_id]);

            let services: ServicesType[] = [];

            for (let i = 0; i < rows.length; i++) {
                const row = rows[i];

                if (row.store_id != store_id) {
                    console.error('store_id does not match');
                    continue;
                }

                const service: ServicesType = {
                    id: row.service_id,
                    name: row.name,
                    activities: [],
                };

                const activityRows = await conn.query('SELECT * FROM store_service_activities WHERE service_id=?', [service.id]);

                for (let j = 0; j < activityRows.length; j++) {
                    const activityRow = activityRows[j];

                    if (activityRow.services_id != row.id) {
                        console.error('services_id does not match');
                        continue;
                    }

                    let users: ActivityUsersType = await getUserIDsByActivityID(store_id, activityRow.activity_id, displayMinimal);

                    const activity: ActivityType = {
                        id: activityRow.activity_id,
                        name: activityRow.name,
                        description: activityRow.description,
                        price: activityRow.price,
                        duration: activityRow.duration,
                        users: users,
                    };

                    service.activities.push(activity);
                }

                services.push(service);
            }

            resolve(services);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getActivity(store_id: StoreID, activity_id: ActivityTypeId, displayMinimal: boolean): Promise<ActivityType> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT * FROM store_service_activities WHERE activity_id=?', [activity_id]);

            let activity: ActivityType;

            if (rows.length >= 1) {
                const row = rows[0];

                if (row.activity_id != activity_id) {
                    reject('activity_id does not match');
                    return;
                }

                let users: ActivityUsersType = await getUserIDsByActivityID(store_id, activity_id, displayMinimal);

                activity = {
                    id: row.activity_id,
                    name: row.name,
                    description: row.description,
                    price: row.price,
                    duration: row.duration,
                    users: users,
                };
            } else {
                reject('no activity found');
                return;
            }

            resolve(activity);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

export { getServices, getActivity, ServicesType, ActivityType, ServicesTypeId, ActivityTypeId };
