import { defaultUserSettings } from '../config';
import logger, { userLogger } from '../logger/logger';
import redisClient from '../redis/init';
import MyRedisKeys from '../redis/keys';
import { User, UserID } from '../user/types';
import { getConnection } from './initDatabase';
import { StoreID, getStore, storeSettingsKeys } from './store';

async function getUser(user_id: UserID): Promise<User> {
    return new Promise(async (resolve, reject) => {
        try {
            //check if user is in redis
            let redisUser = await redisClient.get(MyRedisKeys.userCache(user_id));
            if (redisUser) {
                let user: User = JSON.parse(redisUser);
                resolve(user);
                return;
            }
        } catch (error) {}

        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT * FROM users WHERE user_id=?', [user_id]);

            let row;

            if (rows.length >= 1) {
                row = rows[0];
            } else {
                reject('no user found ' + user_id);
                return;
            }

            if (row.user_id != user_id) {
                reject('user_id does not match');
                return;
            }

            let user: User;

            const store = await getStore(row.store_id);

            let _settings: any = {};

            for (const key of storeSettingsKeys) {
                _settings[key] = row[key];
            }

            let settings: User['settings'] = { ...store.settings }; //{ ...defaultUserSettings };

            for (const _key in settings) {
                const key = _key as keyof User['settings'];

                if (_settings[key] !== undefined && _settings[key] !== null) {
                    if (key === 'calendar_using_primary_calendar') {
                        (settings as any)[key] = _settings[key] === 1;
                    } else {
                        (settings as any)[key] = _settings[key];
                    }
                }
            }

            user = {
                user_id: row.user_id,
                store_id: row.store_id,
                username: row.username,
                lastUpdate: new Date(),
                isOwner: store.owner_user_id == row.user_id,
                settings,
                email: row.email ? row.email : undefined,
            };

            //cache user
            try {
                await redisClient.set(MyRedisKeys.userCache(user_id), JSON.stringify(user), { EX: 1 * 60 });
            } catch (error) {}

            resolve(user);
        } catch (err) {
            userLogger.error(user_id, 'getUser', (err as Error).toString());
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getAllUsersIDsByStoreID(store_id: StoreID): Promise<UserID[]> {
    let conn;
    try {
        conn = await getConnection();
        const rows = await conn.query('SELECT user_id FROM users WHERE store_id=?', [store_id]);

        let userIDs: UserID[] = [];

        for (const row of rows) {
            userIDs.push(row.user_id);
        }

        return userIDs;
    } catch (err) {
        console.log(err);
        throw err;
    } finally {
        if (conn) conn.end();
    }
}

async function getAllUsersIDsAndNamesByStoreID(store_id: StoreID): Promise<{ [key: UserID]: { name: string } }> {
    let conn;
    try {
        conn = await getConnection();
        const rows = await conn.query('SELECT user_id, username,  FROM users WHERE store_id=?', [store_id]);

        let userIDs: { [key: UserID]: { name: string } } = {};

        for (const row of rows) {
            userIDs[row.user_id] = { name: row.username };
        }

        return userIDs;
    } catch (err) {
        console.log(err);
        throw err;
    } finally {
        if (conn) conn.end();
    }
}

export { getUser, getAllUsersIDsByStoreID, getAllUsersIDsAndNamesByStoreID };

/*
CREATE TABLE `users` (
  `user_id` varchar(255) NOT NULL,
  `store_id` varchar(255) NOT NULL,
  `account_name` varchar(255) DEFAULT NULL,
  `username` varchar(255) NOT NULL,
  `password` varchar(255) NOT NULL,
  `calendar_max_future_booking_days` int(11) DEFAULT NULL,
  `calendar_min_earliest_booking_time` int(11) DEFAULT NULL,
  `calendar_primary_calendar_id` varchar(255) DEFAULT NULL,
  `calendar_using_primary_calendar` tinyint(1) DEFAULT NULL,
  `createdAt` datetime NOT NULL,
  `updatedAt` datetime NOT NULL
)
*/
