import { forceGetFreshOpenHours } from '../calendar/openHoursList';
import { clientId, clientSecret } from '../config';
import clearRedisCache from '../redis/clearCache';
import fillCache from '../redis/fillCache';
import { getConnection } from './initDatabase';
import { OAuth2Client } from 'google-auth-library';
import { getUser } from './user';
import calendar_ids from './calendar_ids';

type GoogleTokenStatus = 'OK' | 'PENDING' | 'ERROR' | 'NOPERM';

async function initGoogleAccount(user_id: string, access_token: string, refresh_token: string, expiry_date: Date, google_uuid: string) {
    return new Promise(async (resolve, reject) => {
        let alreadySavedTokens = null;
        try {
            alreadySavedTokens = await getGoogleTokens(user_id);
        } catch (error) {}

        let conn;
        try {
            conn = await getConnection();
            let rows;
            if (!alreadySavedTokens) {
                rows = await conn.query('INSERT INTO user_google_tokens (user_id, access_token, refresh_token, status, expiry_date, google_uuid) VALUES (?, ?, ?, ?, ?, ?)', [
                    user_id,
                    access_token,
                    refresh_token,
                    'PENDING' as GoogleTokenStatus,
                    expiry_date,
                    google_uuid,
                ]);
            } else {
                rows = await conn.query('UPDATE user_google_tokens SET access_token=?, refresh_token=?, status=?, expiry_date=?, google_uuid=? WHERE user_id=?', [
                    access_token,
                    refresh_token,
                    'PENDING' as GoogleTokenStatus,
                    expiry_date,
                    google_uuid,
                    user_id,
                ]);
            }

            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function updateGoogleAccountStatus(user_id: string, status: GoogleTokenStatus) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            let rows;

            rows = await conn.query('UPDATE user_google_tokens SET status=? WHERE user_id=?', [status, user_id]);

            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function unlinkGoogleAccount(user_id: string) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            let rows;

            try {
                let alreadySavedTokens = await getGoogleTokens(user_id);

                const oauth2Client = new OAuth2Client({
                    clientId,
                    clientSecret,
                });

                oauth2Client.setCredentials({
                    access_token: alreadySavedTokens.access_token,
                    refresh_token: alreadySavedTokens.refresh_token,
                    //expiry_date: tokens.expiry_date.getTime(),
                });

                await oauth2Client.revokeToken(alreadySavedTokens.refresh_token);
            } catch (error) {}

            try {
                const rows = await conn.query('SELECT * FROM calendar_ids WHERE user_id=?', [user_id]);

                for (let i = 0; i < rows.length; i++) {
                    const calendar = rows[i];

                    if (calendar.user_id != user_id) {
                        console.error('user_id does not match');
                        continue;
                    }

                    try {
                        await calendar_ids.clearCalendarDatabase(calendar.calendar_id);
                    } catch (error) {}
                }

                await conn.query('UPDATE calendar_ids SET is_disabled=1, sync_token=NULL WHERE user_id=?', [user_id]);

                const user = await getUser(user_id);

                // clear whole redis cache of this store
                await clearRedisCache(user.store_id);

                // update open hours
                await forceGetFreshOpenHours(user.store_id);

                setTimeout(() => {
                    fillCache(user.store_id);
                }, 500);
            } catch (error) {}

            rows = await conn.query('DELETE FROM user_google_tokens WHERE user_id=?', [user_id]);

            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

type GoogleTokens = {
    user_id: string;
    access_token: string;
    refresh_token: string;
    expiry_date: Date;
    google_uuid: string;
    created_at: Date;
    updated_at: Date;
};

async function getGoogleTokens(userId: string): Promise<GoogleTokens> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT * FROM user_google_tokens WHERE user_id=?', [userId]);

            if (rows.length >= 1) {
                const tokens: GoogleTokens = rows[0];

                if (tokens.user_id != userId) {
                    reject('user_id does not match');
                    return;
                }

                resolve(tokens);
            } else {
                reject('no google tokens found for ' + userId);
            }
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function updateAccessToken(userId: string, accessToken: string, expire_date: Date) {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('UPDATE user_google_tokens SET access_token=?, expiry_date=? WHERE user_id=?', [accessToken, expire_date, userId]);
            console.log(rows); //[ {val: 1}, meta: ... ]
            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

async function getUserIds(): Promise<{ user_id: string }[]> {
    return new Promise(async (resolve, reject) => {
        let conn;
        try {
            conn = await getConnection();
            const rows = await conn.query('SELECT user_id FROM user_google_tokens');
            console.log(rows); //[ {val: 1}, meta: ... ]
            resolve(rows);
        } catch (err) {
            console.log(err);
            reject(err);
        } finally {
            if (conn) conn.end();
        }
    });
}

const GoogleTokensDatabase = {
    initGoogleAccount,
    unlinkGoogleAccount,
    getGoogleTokens,
    updateAccessToken,
    getUserIds,
    updateGoogleAccountStatus,
};

export default GoogleTokensDatabase;

/* Database structure:

CREATE TABLE `user_google_tokens` (
  `user_id` varchar(255) NOT NULL,
  `access_token` text NOT NULL,
  `refresh_token` text NOT NULL,
  `expiry_date` timestamp NOT NULL,
  `google_uuid` text NOT NULL,
  `created_at` timestamp NULL DEFAULT current_timestamp(),
  `updated_at` timestamp NULL DEFAULT current_timestamp() ON UPDATE current_timestamp()
)

*/
