import { Request, Response, NextFunction } from 'express';
import app from './express';
import { GLOBAL_calendar_min_earliest_booking_time, publicFolder, publicURL } from '../config';
import { getAccessToken, getOAuthClient, initOauthAccess } from '../calendar/oauthAccess';
import { google } from 'googleapis';
import calendar_ids from '../database/calendar_ids';
import { getFreeNextDays, getFreeTimeOfDay } from '../calendar/getFreeTime';
import { getAllUsersIDsAndNamesByStoreID, getAllUsersIDsByStoreID, getUser } from '../database/user';
import { MasterUser, UserID } from '../user/types';
import { getStore } from '../database/store';
import { getActivity, getServices } from '../database/services';
import { checkIfTimeIsAvailable } from '../calendar/insertEvent';
import MyEvents from '../database/reservedCustomer';
import initCalendar, { removeCalendar } from '../calendar/initCalendar';
import verifyCaptcha from '../captcha';
import sendVerifyMail from '../mail/verify';
import GoogleTokensDatabase from '../database/user_google_tokens';
import { forceGetFreshOpenHours } from '../calendar/openHoursList';
import { scheduleCalendarSync } from '../calendarSync/refreshSync';
import MyRedisKeys from '../redis/keys';
import redisClient from '../redis/init';
import clearRedisCache from '../redis/clearCache';
import { getHighestCalendarMaxFutureBookingDays } from '../calendarSync/openHours';

import fs from 'fs';
import { cancelEmailNotifications } from '../mail/notifyBeforeAppointment';
import logger, { storeLogger, userLogger } from '../logger/logger';

import * as EmailValidator from 'email-validator';
import { log } from 'console';

function getIP(req: Request) {
    return req.headers['x-real-ip'] as string;
}

function activateFileServe() {
    app.get('/embedPopup/script.js', async function (req: Request, res: Response) {
        try {
            // read file publicFolder + '/embedPopup/script.js' by fs.readFile
            const file = await new Promise<Buffer>((resolve, reject) => {
                fs.readFile(publicFolder + '/embedPopup/script.js', (err, data) => {
                    if (err) reject(err);
                    else resolve(data);
                });
            });

            if (!file) throw new Error('file not found');

            if (publicURL === undefined) throw new Error('publicURL in .env not found');

            // replace %PUBLIC_URL% with config.publicURL
            const fileString = file.toString().replace(/%PUBLIC_URL%/g, publicURL);

            // send file as response
            res.setHeader('Content-Type', 'application/javascript');
            res.send(fileString);
        } catch (error) {
            console.log(error);
            res.status(500).send('error 500');
        }
    });
    app.get('/embedPopup/style.css', function (req: Request, res: Response) {
        res.sendFile(publicFolder + '/embedPopup/style.css');
    });

    app.get('/ttest', async function (req: Request, res: Response) {
        res.sendFile(publicFolder + '/index.html');

        let userId = req.cookies['token'] as string | undefined;

        if (userId) {
            try {
                const accessToken = await getAccessToken(userId);
                console.log('accessToken: ' + accessToken);
            } catch (error) {}
        }
    });

    app.post('/getBooking', async function (req: Request, res: Response) {
        const customer_id = req.body.bookingID;
        const ip = getIP(req);

        // check if name is too long
        if (!customer_id || customer_id.length < 32) {
            res.status(400).send(JSON.stringify({ error: 'no valid id given' }));
            logger.warn('REQUEST /getBooking', 'no valid id given', customer_id ? customer_id : '"no id given"', ip);
            return;
        }

        try {
            const event = await MyEvents.getPublicEvent(customer_id);

            if (event && event.time_start > new Date()) {
                res.status(200).send(JSON.stringify({ event: event }));
            } else {
                res.status(404).send(JSON.stringify({ error: 'not found' }));

                logger.warn('/getBooking', 'no valid id given', customer_id ? customer_id : '"no id given"', ip);
            }
        } catch (error) {
            logger.error('REQUEST /getBooking', 'error', (error as Error).toString(), ip);
            res.status(500).send(JSON.stringify({ error: 'unknown error' }));
        }
    });

    app.post('/booking', async function (req: Request, res: Response) {
        // get store_id from body
        const store_id = req.body.store_id;
        const activity_id = req.body.activity_id;
        const date = req.body.time;
        const name = req.body.name;
        const email = req.body.email;
        let message = req.body.message;
        const employee = req.body.employee;
        const captcha = req.body.captcha;
        const ip = getIP(req);

        if (store_id) {
            try {
                const store = await getStore(store_id);

                // check if date is valid
                if (isNaN(Number(date))) {
                    res.status(400).send(JSON.stringify({ error: 'invalid date' }));
                    storeLogger.warn(store_id, 'REQUEST /booking', 'invalid date', date, ip);
                    return;
                }

                // check if name is valid
                if (name === undefined || name === null || name.length < 3) {
                    res.status(400).send(JSON.stringify({ error: 'invalid name' }));
                    storeLogger.warn(store_id, 'REQUEST /booking', 'invalid name', name, ip);
                    return;
                }

                // check if name is too long
                if (name.length > 32) {
                    res.status(400).send(JSON.stringify({ error: 'name too long' }));
                    storeLogger.warn(store_id, 'REQUEST /booking', 'name too long', name, ip);
                    return;
                }

                // check if email is valid
                if (email === undefined || email === null || !EmailValidator.validate(email)) {
                    res.status(400).send(JSON.stringify({ error: 'invalid email' }));
                    storeLogger.warn(store_id, 'REQUEST /booking', 'invalid email', email, ip);
                    return;
                }

                // check if email is too long
                if (email.length > 255) {
                    res.status(400).send(JSON.stringify({ error: 'email too long' }));
                    storeLogger.warn(store_id, 'REQUEST /booking', 'email too long (' + email.length + ')', ip);
                    return;
                }

                // check if message is valid
                if (message === undefined || message === null) {
                    message = '';
                } else if (message.length > 1502) {
                    res.status(400).send(JSON.stringify({ error: 'message too long' }));
                    storeLogger.warn(store_id, 'REQUEST /booking', 'message too long (' + message.length + ')', ip);
                    return;
                }

                try {
                    await verifyCaptcha(captcha, ip);
                } catch (error) {
                    storeLogger.warn(store_id, 'REQUEST /booking', 'captcha error', (error as Error).toString(), ip);
                    res.status(400).send(JSON.stringify({ error: 'captcha' }));
                    return;
                }

                const calendarLockIDs: {
                    [key: string]: {
                        calendar_id: string | undefined;
                        isLocked: boolean;
                        refreshLock: () => void;
                        removeLock: () => void;
                    };
                } = {
                    event: {
                        calendar_id: undefined,
                        isLocked: false,
                        refreshLock: async () => {},
                        removeLock: async () => {},
                    },
                    primary: {
                        calendar_id: undefined,
                        isLocked: false,
                        refreshLock: async () => {},
                        removeLock: async () => {},
                    },
                };

                try {
                    const activity = await getActivity(store_id, activity_id, false);

                    // check if employee is valid
                    if (employee === undefined || employee === null || !activity.users.includes(employee)) {
                        res.status(400).send(JSON.stringify({ error: 'invalid employee' }));
                        storeLogger.warn(store_id, 'REQUEST /booking', 'invalid employee', employee, ip);
                        return;
                    }

                    // start and end time
                    const startTime = new Date(Number(date));
                    const endTime = new Date(Number(date) + activity.duration * 60 * 1000);

                    const user = await getUser(employee);

                    //wait for calendarSyncLock
                    await new Promise(async (resolve, reject) => {
                        let isLocked = true;

                        calendarLockIDs.event.calendar_id = await calendar_ids.getEventCalendarID(user);

                        try {
                            calendarLockIDs.primary.calendar_id = (await calendar_ids.getPrimaryCalendar(user.user_id)).calendar_id;
                        } catch (error) {}

                        while (isLocked) {
                            let hasFoundSomethingLocked = false;

                            for (const key in calendarLockIDs) {
                                const calendarID = calendarLockIDs[key];
                                if (calendarID.calendar_id) {
                                    const lockKey = MyRedisKeys.calendarSyncLock(calendarID.calendar_id);

                                    if (calendarID.isLocked) {
                                        await calendarID.refreshLock();
                                        continue;
                                    }

                                    const lock = await redisClient.set(lockKey, 'locked', { NX: true, EX: 30 });

                                    if (lock) {
                                        calendarID.isLocked = true;
                                        calendarID.refreshLock = async () => {
                                            await redisClient.expire(lockKey, 30);
                                        };
                                        calendarID.removeLock = async () => {
                                            await redisClient.del(lockKey);
                                        };
                                    } else {
                                        hasFoundSomethingLocked = true;
                                    }
                                }
                            }

                            if (hasFoundSomethingLocked) {
                                await new Promise((resolve) => setTimeout(resolve, 500));
                            } else {
                                isLocked = false;
                            }
                        }

                        resolve(undefined);
                    });

                    const lockTimeout = 30; // 30 seconds
                    const lockValue = 'locked';
                    let lockDate = new Date();

                    let lockKeyUnique = MyRedisKeys.userBookingLockUnique(user.user_id, startTime, lockDate);
                    const lockKey = MyRedisKeys.userBookingLock(user.user_id, startTime);

                    const refreshLock = () => {
                        redisClient.expire(lockKeyUnique, lockTimeout);
                    };

                    try {
                        // lock user
                        let lock = await redisClient.set(lockKeyUnique, lockValue, { NX: true, EX: lockTimeout });

                        while (!lock) {
                            lockDate = new Date();

                            lockKeyUnique = MyRedisKeys.userBookingLockUnique(user.user_id, startTime, lockDate);
                            lock = await redisClient.set(lockKeyUnique, lockValue, { NX: true, EX: lockTimeout });

                            if (!lock) {
                                // wait 500ms
                                await new Promise((resolve) => setTimeout(resolve, 500 + Math.random() * 500));
                            }
                            refreshLock();
                        }

                        const splitLength = lockKeyUnique.split(':').length;

                        let isFirstQueue = false;
                        while (!isFirstQueue) {
                            // list all keys
                            const keys = await redisClient.keys(lockKey + ':*');

                            let oldestTime = -1;
                            // check if first in queue
                            for (let i = 0; i < keys.length; i++) {
                                const key = keys[i];

                                const time = Number(key.split(':')[splitLength - 1]);

                                if (time < oldestTime || oldestTime === -1) {
                                    oldestTime = time;
                                }
                            }

                            if (oldestTime === Number(lockKeyUnique.split(':')[splitLength - 1])) {
                                isFirstQueue = true;
                            } else {
                                await new Promise((resolve) => setTimeout(resolve, 2000));
                            }
                            refreshLock();
                        }

                        // check if time is available
                        try {
                            await clearRedisCache(store_id);
                            await checkIfTimeIsAvailable(store, user, startTime, endTime);
                        } catch (error) {
                            console.error(error);
                            userLogger.error(user.user_id, 'REQUEST /booking', 'time not available', (error as Error).toString(), ip);
                            res.status(409).send(JSON.stringify({ status: 'already booked' }));
                            return;
                        }

                        // create event
                        const event = await MyEvents.createEvent(store, user, activity_id, startTime, endTime, email, name, message);

                        userLogger.info(user.user_id, 'REQUEST /booking', 'event created', JSON.stringify(event, null, 2), ip);

                        // send email
                        try {
                            await sendVerifyMail(event);
                        } catch (error) {
                            console.error(error);
                            userLogger.error(user.user_id, 'REQUEST /booking', 'email not sent', (error as Error).toString(), JSON.stringify(event, null, 2), ip);
                            res.status(500).send(JSON.stringify({ error: 'unknown error' }));
                            return;
                        }

                        res.status(200).send(JSON.stringify({ status: 'success' }));
                    } catch (error) {
                        console.log(error);
                        userLogger.error(user.user_id, 'REQUEST /booking', 'unknown error', (error as Error).toString(), ip);
                        res.status(500).send(JSON.stringify({ error: 'unknown error' }));
                    }
                } catch (error) {
                    console.log(error);
                    storeLogger.error(store_id, 'REQUEST /booking', 'unknown error', (error as Error).toString(), ip);
                    res.status(500).send(JSON.stringify({ error: 'unknown error' }));
                } finally {
                    for (const key in calendarLockIDs) {
                        const calendarID = calendarLockIDs[key];
                        if (calendarID.isLocked) {
                            await calendarID.removeLock();
                        }
                    }
                }
            } catch (error) {
                console.log(error);
                storeLogger.error(store_id, 'REQUEST /booking', 'unknown error', (error as Error).toString(), ip);
                res.status(404).send(JSON.stringify({ error: 'store not found' }));
            }
        } else {
            logger.warn('REQUEST /booking', 'no storeId given', ip);
            res.status(400).send(JSON.stringify({ error: 'store not found' }));
        }
    });

    app.post('/services', async function (req: Request, res: Response) {
        // get store_id from body
        const store_id = req.body.store_id;
        const ip = getIP(req);

        if (store_id) {
            try {
                const store = await getStore(store_id);

                try {
                    const services = await getServices(store.store_id, true);
                    const _users = await getAllUsersIDsByStoreID(store.store_id);
                    let users: { [key: string]: { name: string; cmebt: number } } = {};

                    for (let i = 0; i < _users.length; i++) {
                        const user = await getUser(_users[i]);
                        users[user.user_id] = { name: user.username, cmebt: user.settings.calendar_min_earliest_booking_time + GLOBAL_calendar_min_earliest_booking_time };
                    }

                    const highest_calendar_max_future_booking_days = await getHighestCalendarMaxFutureBookingDays(store.store_id);

                    const maxDate = new Date(new Date().getTime() + highest_calendar_max_future_booking_days * 24 * 60 * 60 * 1000);

                    res.status(200).send(JSON.stringify({ services: services, users: users, maxDate: maxDate.getTime(), storeName: store.name }));
                } catch (error) {
                    console.log(error);
                    storeLogger.error(store_id, 'REQUEST /services', 'unknown error', (error as Error).toString(), ip);
                    res.status(500).send('error 500');
                }
            } catch (error) {
                console.log(error);
                logger.warn('REQUEST /services', 'store not found', (error as Error).toString(), ip);
                res.status(404).send('store not found');
            }
        } else {
            logger.warn('REQUEST /services', 'no store id found', ip);
            res.status(400).send('id not found');
        }
    });

    app.post('/freeByActivity', async function (req: Request, res: Response) {
        const store_id = req.body.store_id;
        const dateQuery = Number(req.body.date);
        const activity_id = req.body.activity_id;

        const wishUser = req.body.users;

        const ip = getIP(req);

        if (store_id) {
            try {
                const store = await getStore(store_id);

                try {
                    const activity = await getActivity(store_id, activity_id, false);
                    const serviceDuration = activity.duration;

                    // check if wishUser is an array
                    if (wishUser !== undefined && !Array.isArray(wishUser)) {
                        logger.warn(store_id, 'REQUEST /freeByActivity', 'invalid users', JSON.stringify(wishUser, null, 2), ip);
                        res.status(400).send('invalid users');
                        return;
                    }

                    if (isNaN(dateQuery)) throw new Error('invalid date');

                    let date = new Date(dateQuery || Date.now());

                    let users: UserID[] = [];

                    if (wishUser === undefined || wishUser.length == 0) {
                        users = activity.users;
                    } else {
                        for (let i = 0; i < wishUser.length; i++) {
                            if (wishUser[i] !== null && wishUser[i] !== undefined) {
                                const userId = wishUser[i].toString();
                                // check if user contains in activity.users
                                if (activity.users.includes(userId)) {
                                    users.push(userId);
                                }
                            }
                        }
                    }

                    const freeTime = await getFreeTimeOfDay(store_id, users, serviceDuration, date);

                    // convert all dates to timestamps
                    for (const key in freeTime) {
                        const time = freeTime[key].events;
                        for (let i = 0; i < time.length; i++) {
                            const timeSlot = time[i];
                            (timeSlot.start as any) = timeSlot.start.getTime();
                            (timeSlot.end as any) = timeSlot.end.getTime();
                        }
                    } /* [{"userId":"ayg2xg0w0vwswbl1c6w28","events":[{"start":1704964831091,"end":1704966300000},{"start":1704969900000,"end":1704970800000},{"start":1704978900000,"end":1704980700000},{"start":1704988800000,"end":1704990600000},{"start":1704994200000,"end":1704997800000},{"start":1705001400000,"end":1705005000000}]}] */

                    res.status(200).send(JSON.stringify({ events: freeTime }));
                } catch (error) {
                    console.log(error);
                    storeLogger.error(store_id, 'REQUEST /freeByActivity', 'error', (error as Error).toString(), ip);
                    res.status(500).send('error 500');
                }
            } catch (error) {
                console.log(error);
                logger.warn('REQUEST /freeByActivity', 'store not found', (error as Error).toString(), ip);
                res.status(400).send('no store found');
            }
        } else {
            res.status(400).send('no store found');
            logger.warn('REQUEST /freeByActivity', 'no store found', ip);
        }
    });

    app.post('/freeNextDaysByActivity', async function (req: Request, res: Response) {
        const store_id = req.body.store_id;
        const dateQuery = Number(req.body.date);
        const activity_id = req.body.activity_id;

        const wishUser = req.body.users;

        const ip = getIP(req);

        if (store_id) {
            try {
                const store = await getStore(store_id);

                try {
                    const activity = await getActivity(store_id, activity_id, false);
                    const serviceDuration = activity.duration;

                    // check if wishUser is an array
                    if (wishUser !== undefined && !Array.isArray(wishUser)) {
                        res.status(400).send('invalid users');
                        logger.warn(store_id, 'REQUEST /freeNextDaysByActivity', 'invalid users', JSON.stringify(wishUser, null, 2), ip);
                        return;
                    }

                    if (isNaN(dateQuery)) {
                        res.status(400).send('invalid date');
                        logger.warn(store_id, 'REQUEST /freeNextDaysByActivity', 'invalid date', dateQuery.toString(), ip);
                        return;
                    }
                    if (isNaN(serviceDuration) || serviceDuration > 24 * 60) {
                        res.status(400).send('invalid service duration');
                        logger.warn(store_id, 'REQUEST /freeNextDaysByActivity', 'invalid service duration', serviceDuration.toString(), ip);
                        return;
                    }

                    // convert dateQueryNumber to date
                    let date = new Date(dateQuery || Date.now());

                    let users: UserID[] = [];

                    if (wishUser === undefined || wishUser.length == 0) {
                        users = activity.users;
                    } else {
                        for (let i = 0; i < wishUser.length; i++) {
                            if (wishUser[i] !== null && wishUser[i] !== undefined) {
                                const userId = wishUser[i].toString();
                                // check if user contains in activity.users
                                if (activity.users.includes(userId)) {
                                    users.push(userId);
                                }
                            }
                        }
                    }

                    const _freeDays = await getFreeNextDays(store.store_id, users, serviceDuration, date);
                    let freeDays: number[] = []; // timestamp

                    // convert all dates to timestamps
                    for (let i = 0; i < _freeDays.length; i++) {
                        freeDays.push(_freeDays[i].getTime());
                    }

                    res.status(200).send(JSON.stringify({ freeDays: freeDays }));
                } catch (error) {
                    console.log(error);
                    res.status(500).send('error 500');
                    storeLogger.error(store_id, 'REQUEST /freeNextDaysByActivity', 'error', (error as Error).toString(), ip);
                }
            } catch (error) {
                console.log(error);
                res.status(404).send('store not found');
                logger.warn('REQUEST /freeNextDaysByActivity', 'store not found', (error as Error).toString(), ip);
            }
        } else {
            res.status(404).send('store not found');
            logger.warn('REQUEST /freeNextDaysByActivity', 'no storeId given', ip);
        }
    });

    // for google api webhook
    app.post('/api/v1/notification/:eventId', async function (req: Request, res: Response) {
        const eventId = req.params.eventId;

        const headers = req.headers;
        const token = headers['x-goog-channel-token'] as string;

        const ip = getIP(req);

        if (!token) {
            console.log(headers);
            res.status(404).send('Not found');

            logger.warn('REQUEST /api/v1/notification', 'no token found', ip);
            return;
        }

        if (token !== process.env.GOOGLE_WEBHOOK_SECRET) {
            console.log(headers);
            res.status(404).send('Not found');

            logger.warn('REQUEST /api/v1/notification', 'invalid token', ip);
            return;
        }

        const state = headers['x-goog-resource-state'] as string;
        const channel_id = headers['x-goog-channel-id'] as string;
        const channel_expiration = headers['x-goog-channel-expiration'] as string;
        const resource_id = headers['x-goog-resource-id'] as string;

        if (state === 'sync') {
            res.status(200).send('success');
            return;
        }

        logger.info('REQUEST /api/v1/notification', 'state', state, 'channel_id', channel_id, 'channel_expiration', channel_expiration, 'resource_id', resource_id, ip);

        try {
            const channelIDExists = await scheduleCalendarSync(eventId, channel_id);

            if (!channelIDExists) {
                res.status(404).send('Not found');
                logger.warn('REQUEST /api/v1/notification', 'channel_id not found', ip);
                return;
            }

            res.status(200).send('success');
        } catch (error) {
            console.error('notificationError', error);
            logger.error('REQUEST /api/v1/notification', 'eventId:', eventId, 'error', (error as Error).toString(), ip);
            res.status(500).send('error 500');
        }
    });

    app.post('/api/v1/changeFutureBookingDays', async function (req: Request, res: Response) {
        let storeId = req.body.storeId as string;
        let pass = req.body.pass as string;

        const ip = getIP(req);

        if (pass !== process.env.BACKEND_PASSPHRASE) {
            res.status(404).send('Not found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'invalid passphrase', ip);
            return;
        }

        if (!storeId) {
            res.status(400).send('no storeId found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'no storeId given', ip);
            return;
        }

        try {
            try {
                await clearRedisCache(storeId);
                await forceGetFreshOpenHours(storeId);
            } catch (error) {}

            res.status(200).send('success');
        } catch (error) {
            storeLogger.error(storeId, 'REQUEST /api/v1/changeFutureBookingDays', 'error', (error as Error).toString(), ip);
            res.status(404).send('unknown error');
        }
    });

    app.post('/api/v1/changedSettings', async function (req: Request, res: Response) {
        let storeId = req.body.storeId as string;
        let pass = req.body.pass as string;

        const ip = getIP(req);

        if (pass !== process.env.BACKEND_PASSPHRASE) {
            res.status(404).send('Not found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'invalid passphrase', ip);
            return;
        }

        if (!storeId) {
            res.status(400).send('no userId found');
            return;
        }

        try {
            try {
                await clearRedisCache(storeId);
                await forceGetFreshOpenHours(storeId);
            } catch (error) {}

            res.status(200).send('success');
        } catch (error) {
            console.log(error);
            storeLogger.error(storeId, 'REQUEST /api/v1/changedSettings', 'error', (error as Error).toString(), ip);
            res.status(404).send('user not found');
        }
    });

    app.post('/api/v1/addGoogleAccount', async function (req: Request, res: Response) {
        let userId = req.body.userId as string;
        let accessToken = req.body.accessToken as string;
        let refreshToken = req.body.refreshToken as string;
        let googleUuid = req.body.sub as string;
        let pass = req.body.pass as string;

        const ip = getIP(req);

        if (pass !== process.env.BACKEND_PASSPHRASE) {
            res.status(404).send('Not found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'invalid passphrase', ip);
            return;
        }

        if (!userId) {
            res.status(400).send('no userId found');
            logger.error('REQUEST /api/v1/addGoogleAccount', 'no userId given', ip);
            return;
        }

        let user = undefined;

        try {
            user = await getUser(userId);
        } catch (error) {
            console.log(error);
            res.status(404).send('user not found');
            logger.error('REQUEST /api/v1/addGoogleAccount', 'user not found', ip);
            return;
        }

        if (!accessToken) {
            res.status(400).send('no accessToken found');
            userLogger.error(user.user_id, 'REQUEST /api/v1/addGoogleAccount', 'no accessToken given', ip);
            return;
        }

        if (!refreshToken) {
            const oauth2Client = await getOAuthClient(accessToken);

            await oauth2Client.revokeToken(accessToken);

            res.status(400).send('no refreshToken found');
            userLogger.error(user.user_id, 'REQUEST /api/v1/addGoogleAccount', 'no refreshToken given', ip);
            return;
        }

        let expireDate = new Date(new Date().getTime() + 1000 * 60 * 30);

        try {
            await initOauthAccess(userId, refreshToken, accessToken, expireDate, googleUuid);

            res.status(200).send('success');

            try {
                await forceGetFreshOpenHours(user.store_id);
            } catch (error) {}
        } catch (error) {
            try {
                const oauth2Client = await getOAuthClient(accessToken);
                await oauth2Client.revokeToken(accessToken);
            } catch (error) {}

            userLogger.error(user.user_id, 'REQUEST /api/v1/addGoogleAccount', 'error', (error as Error).toString(), ip);
            res.status(500).send('error 500');
        }
    });

    app.post('/api/v1/removeGoogleAccount', async function (req: Request, res: Response) {
        let userId = req.body.userId as string;
        let pass = req.body.pass as string;

        const ip = getIP(req);

        if (pass !== process.env.BACKEND_PASSPHRASE) {
            res.status(404).send('Not found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'invalid passphrase', ip);
            return;
        }

        if (!userId) {
            res.status(400).send('no userId found');
            logger.error('REQUEST /api/v1/removeGoogleAccount', 'no userId given', ip);
            return;
        }

        try {
            let user;
            try {
                user = await getUser(userId);
            } catch (error) {
                logger.error('REQUEST /api/v1/removeGoogleAccount', 'user not found', ip);
                res.status(404).send('user not found');
                return;
            }

            await GoogleTokensDatabase.unlinkGoogleAccount(userId);

            res.status(200).send('success');

            try {
                await forceGetFreshOpenHours(user.store_id);
            } catch (error) {}
        } catch (error) {
            res.status(500).send('error 500');
            logger.error('REQUEST /api/v1/removeGoogleAccount', 'error', (error as Error).toString(), ip);
        }
    });

    app.post('/api/v1/addUser', async function (req: Request, res: Response) {
        let userId = req.body.userId as string;
        let pass = req.body.pass as string;

        const ip = getIP(req);

        if (pass !== process.env.BACKEND_PASSPHRASE) {
            res.status(404).send('Not found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'invalid passphrase', ip);
            return;
        }

        if (!userId) {
            res.status(400).send('no userId found');
            logger.error('REQUEST /api/v1/addUser', 'no userId given', ip);
            return;
        }

        try {
            const user = await getUser(userId);

            try {
                await initCalendar(user.user_id);

                res.status(200).send('success');
            } catch (error) {
                console.log(error);
                userLogger.error(user.user_id, 'REQUEST /api/v1/addUser', 'error', (error as Error).toString(), ip);
                res.status(500).send('error 500');
            }
        } catch (error) {
            console.log(error);
            logger.error('REQUEST /api/v1/addUser', 'user not found', ip);
            res.status(404).send('user not found');
        }
    });

    app.post('/api/v1/removeUser', async function (req: Request, res: Response) {
        let userId = req.body.userId as string;
        let pass = req.body.pass as string;

        const ip = getIP(req);

        if (pass !== process.env.BACKEND_PASSPHRASE) {
            res.status(404).send('Not found');
            logger.warn('REQUEST /api/v1/changeFutureBookingDays', 'invalid passphrase', ip);
            return;
        }

        if (!userId) {
            res.status(400).send('no userId found');
            logger.error('REQUEST /api/v1/removeUser', 'no userId given', ip);
            return;
        }

        try {
            const user = await getUser(userId);

            try {
                await removeCalendar(user);

                res.status(200).send('success');
            } catch (error) {
                console.log(error);
                userLogger.error(user.user_id, 'REQUEST /api/v1/removeUser', 'error', (error as Error).toString(), ip);
                res.status(500).send('error 500');
            }
        } catch (error) {
            console.log(error);
            logger.error('REQUEST /api/v1/removeUser', 'user not found', ip);
            res.status(404).send('user not found');
        }
    });

    app.get('/appointment/verify/:customer_id', async function (req: Request, res: Response) {
        const customer_id = req.params.customer_id;

        const ip = getIP(req);

        if (customer_id) {
            try {
                const event = await MyEvents.verifyEvent(customer_id);

                if (event) {
                    userLogger.info(event.user_id, 'REQUEST /appointment/verify', 'event verified', ip);
                    res.send(JSON.stringify({ status: 'success' }));
                } else {
                    logger.warn('REQUEST /appointment/verify', 'event not found', customer_id, ip);
                    res.status(404).sendFile(publicFolder + '/404.html');
                }
            } catch (error) {
                console.log(error);
                logger.error('REQUEST /appointment/verify', 'error', customer_id, (error as Error).toString(), ip);
                res.status(500).sendFile(publicFolder + '/error.html');
            }
        } else {
            logger.warn('REQUEST /appointment/verify', 'no customer_id given', ip);
            res.status(400).sendFile(publicFolder + '/error.html');
        }
    });

    app.get('/appointment/cancel/:customer_id', async function (req: Request, res: Response) {
        const customer_id = req.params.customer_id;

        const ip = getIP(req);

        if (customer_id) {
            try {
                const event = await MyEvents.cancelEvent(customer_id);

                if (event) {
                    userLogger.info(event.user_id, 'REQUEST /appointment/cancel', 'event canceled', ip);
                    res.send(JSON.stringify({ status: 'success' }));
                } else {
                    res.status(404).sendFile(publicFolder + '/404.html');
                    logger.warn('REQUEST /appointment/cancel', 'event not found', customer_id, ip);
                }
            } catch (error) {
                console.log(error);
                logger.error('REQUEST /appointment/cancel', 'error', customer_id, (error as Error).toString(), ip);
                res.status(500).sendFile(publicFolder + '/error.html');
            }
        } else {
            logger.warn('REQUEST /appointment/cancel', 'no customer_id given', ip);
            res.status(400).sendFile(publicFolder + '/error.html');
        }
    });

    app.get('/appointment/cancelNotification/:customer_id', async function (req: Request, res: Response) {
        const customer_id = req.params.customer_id;

        const ip = getIP(req);

        if (customer_id) {
            try {
                const event = await cancelEmailNotifications(customer_id);

                if (event) {
                    userLogger.info('REQUEST /appointment/cancelNotification', 'notifications canceled', customer_id, ip);
                    res.send(JSON.stringify({ status: 'success' }));
                } else {
                    res.status(404).sendFile(publicFolder + '/404.html');
                    logger.warn('REQUEST /appointment/cancelNotification', 'event not found', customer_id, ip);
                }
            } catch (error) {
                console.log(error);
                res.status(500).sendFile(publicFolder + '/error.html');
                logger.error('REQUEST /appointment/cancelNotification', 'error', customer_id, (error as Error).toString(), ip);
            }
        } else {
            res.status(400).sendFile(publicFolder + '/error.html');
            logger.warn('REQUEST /appointment/cancelNotification', 'no customer_id given', ip);
        }
    });

    //on 404
    app.use(function (req: Request, res: Response, next: NextFunction) {
        res.status(404).sendFile(publicFolder + '/404.html');
        logger.warn('REQUEST', '404', req.url, getIP(req));
    });
}

export default activateFileServe;

function printDate(str: string, date: Date) {
    console.log(
        str,
        date.toLocaleDateString('de-DE', {
            day: '2-digit',
            month: '2-digit',
            year: 'numeric',
            hour: '2-digit',
            minute: '2-digit',
            second: '2-digit',
        })
    );
}
